@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -45,7 +45,7 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * JSON Web Tokens Signing and Encryption
+ * JSON Object Signing and Encryption
  *
  * @Author: Hans Zandbelt - hzandbelt@pingidentity.com
  */
@@ -53,8 +53,8 @@
 #ifndef _APR_JOSE_H_
 #define _APR_JOSE_H_
 
+#include <stdint.h>
 #include "apr_pools.h"
-#include "apr_time.h"
 #include "apr_tables.h"
 #include "apr_hash.h"
 #include "apr_strings.h"
@@ -63,6 +63,25 @@
 
 #define APR_JWT_CLAIM_TIME_EMPTY -1
 
+#define APR_JWT_ERROR_TEXT_LENGTH    200
+#define APR_JWT_ERROR_SOURCE_LENGTH   80
+#define APR_JWT_ERROR_FUNCTION_LENGTH 80
+
+/* struct for returning errors to the caller */
+typedef struct {
+	char source[JSON_ERROR_SOURCE_LENGTH];
+	int line;
+	char function[APR_JWT_ERROR_FUNCTION_LENGTH];
+	char text[APR_JWT_ERROR_TEXT_LENGTH];
+} apr_jwt_error_t;
+
+void _apr_jwt_error_set(apr_jwt_error_t *, const char *, const int,
+		const char *, const char *msg, ...);
+#define apr_jwt_error(err, msg, ...) _apr_jwt_error_set(err, __FILE__, __LINE__, __FUNCTION__, msg, ##__VA_ARGS__)
+#define apr_jwt_error_openssl(err, msg, ...) _apr_jwt_error_set(err, __FILE__, __LINE__, __FUNCTION__, "%s() failed: %s", msg, ERR_error_string(ERR_get_error(), NULL), ##__VA_ARGS__)
+
+#define apr_jwt_e2s(pool, err) apr_psprintf(pool, "[%s:%d: %s]: %s\n", err.source, err.line, err.function, err.text)
+
 /*
  * JSON Web Token handling
  */
@@ -96,9 +115,9 @@ typedef struct apr_jwt_payload_t {
 	/* JWT "sub" claim value; subject/principal */
 	char *sub;
 	/* parsed JWT "exp" claim value; token expiry */
-	apr_time_t exp;
+	json_int_t exp;
 	/* parsed JWT "iat" claim value; issued-at timestamp */
-	apr_time_t iat;
+	json_int_t iat;
 } apr_jwt_payload_t;
 
 /* parsed JWT signature */
@@ -122,27 +141,34 @@ typedef struct apr_jwt_t {
 } apr_jwt_t;
 
 /* helpers */
-typedef apr_byte_t (*apr_jose_is_supported_function_t)(apr_pool_t *, const char *);
+typedef apr_byte_t (*apr_jose_is_supported_function_t)(apr_pool_t *,
+		const char *);
 apr_byte_t apr_jwt_array_has_string(apr_array_header_t *haystack,
 		const char *needle);
 int apr_jwt_base64url_encode(apr_pool_t *pool, char **dst, const char *src,
 		int src_len, int padding);
 int apr_jwt_base64url_decode(apr_pool_t *pool, char **dst, const char *src,
 		int padding);
-const char *apr_jwt_header_to_string(apr_pool_t *pool, const char *s_json);
+const char *apr_jwt_header_to_string(apr_pool_t *pool, const char *s_json,
+		apr_jwt_error_t *err);
 
 /* return a string claim value from a JSON Web Token */
-apr_byte_t apr_jwt_get_string(apr_pool_t *pool, apr_jwt_value_t *value,
-		const char *claim_name, char **result);
-/* parse a string in to a JSON Web Token struct */
+apr_byte_t apr_jwt_get_string(apr_pool_t *pool, json_t *json,
+		const char *claim_name, apr_byte_t is_mandatory, char **result,
+		apr_jwt_error_t *err);
+/* parse a string into a JSON Web Token struct and (optionally) decrypt it */
 apr_byte_t apr_jwt_parse(apr_pool_t *pool, const char *s_json,
-		apr_jwt_t **j_jwt, apr_hash_t *private_keys, const char *shared_key);
+		apr_jwt_t **j_jwt, apr_hash_t *keys, apr_jwt_error_t *err);
 /* destroy resources allocated for JWT */
 void apr_jwt_destroy(apr_jwt_t *);
 
 /* exported for the purpose of the test suite */
-apr_array_header_t *apr_jwt_compact_deserialize(apr_pool_t *pool, const char *str);
-apr_byte_t apr_jwt_parse_header(apr_pool_t *pool, const char *s_header, apr_jwt_header_t *header);
+apr_byte_t apr_jwt_header_parse(apr_pool_t *pool, const char *s_json,
+		apr_array_header_t **unpacked, apr_jwt_header_t *header,
+		apr_jwt_error_t *err);
+
+/* return the JWK type for the JWT signature verification */
+const char *apr_jwt_signature_to_jwk_type(apr_pool_t *pool, apr_jwt_t *jwt);
 
 /*
  * JSON Web Key handling
@@ -154,6 +180,8 @@ typedef enum apr_jwk_type_e {
 	APR_JWK_KEY_RSA,
 	/* EC JWT key type */
 	APR_JWK_KEY_EC,
+	/* oct JWT key type */
+	APR_JWK_KEY_OCT,
 } apr_jwk_type_e;
 
 /* parsed RSA JWK key */
@@ -184,26 +212,43 @@ typedef struct apr_jwk_key_ec_t {
 	int y_len;
 } apr_jwk_key_ec_t;
 
+/* parsed oct JWK key */
+typedef struct apr_jwk_key_oct_t {
+	/* k octets */
+	unsigned char *k;
+	/* length of k */
+	int k_len;
+} apr_jwk_key_oct_t;
+
 /* parsed JWK key */
 typedef struct apr_jwk_t {
-	/* parsed JWK/JSON value */
-	apr_jwt_value_t value;
+	/* key identifier */
+	char *kid;
 	/* type of JWK key */
 	apr_jwk_type_e type;
 	/* union/pointer to parsed JWK key */
 	union {
 		apr_jwk_key_rsa_t *rsa;
 		apr_jwk_key_ec_t *ec;
+		apr_jwk_key_oct_t *oct;
 	} key;
 } apr_jwk_t;
 
 /* parse a JSON representation in to a JSON Web Key struct (also storing the string representation */
 apr_byte_t apr_jwk_parse_json(apr_pool_t *pool, json_t *j_json,
-		const char *s_json, apr_jwk_t **j_jwk);
-/* convert the RSA public key in a PEM formatted file with an X.509 cert in to an RSA JWK */
-apr_byte_t apr_jwk_pem_to_json(apr_pool_t *pool, const char *filename, char **jwk, char**kid);
-/* convert the RSA private key in a PEM formatted file in to an RSA JWK */
-apr_byte_t apr_jwk_private_key_to_rsa_jwk(apr_pool_t *pool, const char *filename, char **jwk, char**kid);
+		apr_jwk_t **j_jwk, apr_jwt_error_t *err);
+/* parse a symmetric key in to a JSON Web Key (oct) struct */
+apr_byte_t apr_jwk_parse_symmetric_key(apr_pool_t *pool,
+		const unsigned char *key, unsigned int key_len, apr_jwk_t **j_jwk,
+		apr_jwt_error_t *err);
+/* parse an RSA private key from a PEM formatted file */
+apr_byte_t apr_jwk_parse_rsa_private_key(apr_pool_t *pool, const char *filename,
+		apr_jwk_t **j_jwk, apr_jwt_error_t *err);
+/* parse an RSA public key from a PEM formatted file */
+apr_byte_t apr_jwk_parse_rsa_public_key(apr_pool_t *pool, const char *filename,
+		apr_jwk_t **j_jwk, apr_jwt_error_t *err);
+apr_byte_t apr_jwk_to_json(apr_pool_t *pool, apr_jwk_t *jwk, char **s_json,
+		apr_jwt_error_t *err);
 
 /*
  * JSON Web Token Signature handling
@@ -220,20 +265,19 @@ apr_byte_t apr_jws_signature_is_hmac(apr_pool_t *pool, apr_jwt_t *jwt);
 apr_byte_t apr_jws_signature_is_rsa(apr_pool_t *pool, apr_jwt_t *jwt);
 /* check if the signature on a JWT is of type Elliptic Curve */
 apr_byte_t apr_jws_signature_is_ec(apr_pool_t *pool, apr_jwt_t *jwt);
-/* verify the HMAC signature on a JWT */
-apr_byte_t apr_jws_verify_hmac(apr_pool_t *pool, apr_jwt_t *jwt,
-		const char *key, const unsigned int key_len);
-/* verify the RSA signature on a JWT */
-apr_byte_t apr_jws_verify_rsa(apr_pool_t *pool, apr_jwt_t *jwt, apr_jwk_t *jwk);
-/* verify the Elliptic Curve signature on a JWT */
-apr_byte_t apr_jws_verify_ec(apr_pool_t *pool, apr_jwt_t *jwt, apr_jwk_t *jwk);
+
+/* verify the signature on a JWT */
+apr_byte_t apr_jws_verify(apr_pool_t *pool, apr_jwt_t *jwt, apr_hash_t *keys,
+		apr_jwt_error_t *err);
 
 /* hash byte sequence */
 apr_byte_t apr_jws_hash_bytes(apr_pool_t *pool, const char *s_digest,
-		const unsigned char *input, unsigned int input_len, unsigned char **output, unsigned int *output_len);
+		const unsigned char *input, unsigned int input_len,
+		unsigned char **output, unsigned int *output_len, apr_jwt_error_t *err);
 /* hash a string */
 apr_byte_t apr_jws_hash_string(apr_pool_t *pool, const char *alg,
-		const char *msg, char **hash, unsigned int *hash_len);
+		const char *msg, char **hash, unsigned int *hash_len,
+		apr_jwt_error_t *err);
 /* length of hash */
 int apr_jws_hash_length(const char *alg);
 
@@ -251,6 +295,10 @@ apr_array_header_t *apr_jwe_supported_encryptions(apr_pool_t *pool);
 apr_byte_t apr_jwe_encryption_is_supported(apr_pool_t *pool, const char *enc);
 
 apr_byte_t apr_jwe_is_encrypted_jwt(apr_pool_t *pool, apr_jwt_header_t *hdr);
-apr_byte_t apr_jwe_decrypt_jwt(apr_pool_t *pool, apr_jwt_header_t *header, apr_array_header_t *unpacked, apr_hash_t *private_keys, const char *shared_key, char **decrypted);
+apr_byte_t apr_jwe_decrypt_jwt(apr_pool_t *pool, apr_jwt_header_t *header,
+		apr_array_header_t *unpacked, apr_hash_t *keys, char **decrypted,
+		apr_jwt_error_t *err);
+
+apr_byte_t apr_jwt_memcmp(const void *in_a, const void *in_b, size_t len);
 
 #endif /* _APR_JOSE_H_ */
