@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -45,7 +45,7 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * JSON Web Signatures handling
+ * JSON Web Key handling
  *
  * @Author: Hans Zandbelt - hzandbelt@pingidentity.com
  */
@@ -53,46 +53,58 @@
 #include <openssl/pem.h>
 #include <openssl/evp.h>
 #include <openssl/rsa.h>
+#include <openssl/err.h>
 
 #include "apr_jose.h"
 
 /*
  * parse an RSA JWK in raw format (n,e,d)
  */
-static apr_byte_t apr_jwk_parse_rsa_raw(apr_pool_t *pool, apr_jwt_value_t *jwk,
-		apr_jwk_key_rsa_t **jwk_key_rsa) {
+static apr_byte_t apr_jwk_parse_rsa_raw(apr_pool_t *pool, json_t *json,
+		apr_jwk_key_rsa_t **jwk_key_rsa, apr_jwt_error_t *err) {
 
 	/* allocate space */
 	*jwk_key_rsa = apr_pcalloc(pool, sizeof(apr_jwk_key_rsa_t));
 	apr_jwk_key_rsa_t *key = *jwk_key_rsa;
 
-	/* parse the modulus */
+	/* parse the mandatory modulus */
 	char *s_modulus = NULL;
-	apr_jwt_get_string(pool, jwk, "n", &s_modulus);
-	if (s_modulus == NULL)
+	if (apr_jwt_get_string(pool, json, "n", TRUE, &s_modulus, err) == FALSE)
 		return FALSE;
 
-	/* parse the modulus size */
+	/* base64url decode the modulus and get its size */
 	key->modulus_len = apr_jwt_base64url_decode(pool, (char **) &key->modulus,
 			s_modulus, 1);
+	if (key->modulus_len <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_decode of modulus failed");
+		return FALSE;
+	}
 
-	/* parse the exponent */
+	/* parse the mandatory exponent */
 	char *s_exponent = NULL;
-	apr_jwt_get_string(pool, jwk, "e", &s_exponent);
-	if (s_exponent == NULL)
+	if (apr_jwt_get_string(pool, json, "e", TRUE, &s_exponent, err) == FALSE)
 		return FALSE;
 
-	/* parse the exponent size */
+	/* base64url decode the exponent and get its size */
 	key->exponent_len = apr_jwt_base64url_decode(pool, (char **) &key->exponent,
 			s_exponent, 1);
+	if (key->exponent_len <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_decode of exponent failed");
+		return FALSE;
+	}
 
-	/* parse the private exponent */
+	/* parse the optional private exponent */
 	char *s_private_exponent = NULL;
-	apr_jwt_get_string(pool, jwk, "d", &s_private_exponent);
+	apr_jwt_get_string(pool, json, "d", FALSE, &s_private_exponent, NULL);
 	if (s_private_exponent != NULL) {
-		/* parse the private exponent size */
+		/* base64url decode the private exponent and get its size */
 		key->private_exponent_len = apr_jwt_base64url_decode(pool,
 				(char **) &key->private_exponent, s_private_exponent, 1);
+		if (key->private_exponent_len <= 0) {
+			apr_jwt_error(err,
+					"apr_jwt_base64url_decode of private exponent failed");
+			return FALSE;
+		}
 	}
 
 	/* that went well */
@@ -104,7 +116,8 @@ static apr_byte_t apr_jwk_parse_rsa_raw(apr_pool_t *pool, apr_jwt_value_t *jwk,
  * by "input" to a JSON Web Key object
  */
 static apr_byte_t apr_jwk_rsa_bio_to_key(apr_pool_t *pool, BIO *input,
-		apr_jwk_key_rsa_t **jwk_key_rsa, int is_private_key) {
+		apr_jwk_key_rsa_t **jwk_key_rsa, int is_private_key,
+		apr_jwt_error_t *err) {
 
 	X509 *x509 = NULL;
 	EVP_PKEY *pkey = NULL;
@@ -112,15 +125,21 @@ static apr_byte_t apr_jwk_rsa_bio_to_key(apr_pool_t *pool, BIO *input,
 
 	if (is_private_key) {
 		/* get the private key struct from the BIO */
-		if ((pkey = PEM_read_bio_PrivateKey(input, NULL, NULL, NULL)) == NULL)
+		if ((pkey = PEM_read_bio_PrivateKey(input, NULL, NULL, NULL)) == NULL) {
+			apr_jwt_error_openssl(err, "PEM_read_bio_PrivateKey");
 			goto end;
+		}
 	} else {
 		/* read the X.509 struct */
-		if ((x509 = PEM_read_bio_X509_AUX(input, NULL, NULL, NULL)) == NULL)
+		if ((x509 = PEM_read_bio_X509_AUX(input, NULL, NULL, NULL)) == NULL) {
+			apr_jwt_error_openssl(err, "PEM_read_bio_X509_AUX");
 			goto end;
+		}
 		/* get the public key struct from the X.509 struct */
-		if ((pkey = X509_get_pubkey(x509)) == NULL)
+		if ((pkey = X509_get_pubkey(x509)) == NULL) {
+			apr_jwt_error_openssl(err, "X509_get_pubkey");
 			goto end;
+		}
 	}
 
 	/* allocate space */
@@ -129,8 +148,10 @@ static apr_byte_t apr_jwk_rsa_bio_to_key(apr_pool_t *pool, BIO *input,
 
 	/* get the RSA key from the public key struct */
 	RSA *rsa = EVP_PKEY_get1_RSA(pkey);
-	if (rsa == NULL)
+	if (rsa == NULL) {
+		apr_jwt_error_openssl(err, "EVP_PKEY_get1_RSA");
 		goto end;
+	}
 
 	/* convert the modulus bignum in to a key/len */
 	key->modulus_len = BN_num_bytes(rsa->n);
@@ -149,6 +170,8 @@ static apr_byte_t apr_jwk_rsa_bio_to_key(apr_pool_t *pool, BIO *input,
 		BN_bn2bin(rsa->d, key->private_exponent);
 	}
 
+	RSA_free(rsa);
+
 	rv = TRUE;
 
 end:
@@ -164,19 +187,35 @@ end:
 /*
  * parse an RSA JWK in X.509 format (x5c)
  */
-static apr_byte_t apr_jwk_parse_rsa_x5c(apr_pool_t *pool, apr_jwk_t *jwk) {
+static apr_byte_t apr_jwk_parse_rsa_x5c(apr_pool_t *pool, json_t *json,
+		apr_jwk_t *jwk, apr_jwt_error_t *err) {
 
 	apr_byte_t rv = FALSE;
 
 	/* get the "x5c" array element from the JSON object */
-	json_t *v = json_object_get(jwk->value.json, "x5c");
-	if ((v == NULL) || (!json_is_array(v)))
+	json_t *v = json_object_get(json, "x5c");
+	if (v == NULL) {
+		apr_jwt_error(err, "JSON key \"%s\" could not be found", "x5c");
 		return FALSE;
+	}
+	if (!json_is_array(v)) {
+		apr_jwt_error(err,
+				"JSON key \"%s\" was found but its value is not a JSON array",
+				"x5c");
+		return FALSE;
+	}
 
 	/* take the first element of the array */
 	v = json_array_get(v, 0);
-	if ((v == NULL) || (!json_is_string(v)))
+	if (v == NULL) {
+		apr_jwt_error(err, "first element in JSON array is \"null\"");
 		return FALSE;
+	}
+	if (!json_is_string(v)) {
+		apr_jwt_error(err, "first element in array is not a JSON string");
+		return FALSE;
+	}
+
 	const char *s_x5c = json_string_value(v);
 
 	/* PEM-format it */
@@ -192,16 +231,19 @@ static apr_byte_t apr_jwk_parse_rsa_x5c(apr_pool_t *pool, apr_jwk_t *jwk) {
 	BIO *input = NULL;
 
 	/* put it in BIO memory */
-	if ((input = BIO_new(BIO_s_mem())) == NULL)
+	if ((input = BIO_new(BIO_s_mem())) == NULL) {
+		apr_jwt_error_openssl(err, "memory allocation BIO_new/BIO_s_mem");
 		return FALSE;
+	}
 
 	if (BIO_puts(input, s) <= 0) {
 		BIO_free(input);
+		apr_jwt_error_openssl(err, "BIO_puts");
 		return FALSE;
 	}
 
 	/* do the actual parsing */
-	rv = apr_jwk_rsa_bio_to_key(pool, input, &jwk->key.rsa, FALSE);
+	rv = apr_jwk_rsa_bio_to_key(pool, input, &jwk->key.rsa, FALSE, err);
 
 	BIO_free(input);
 
@@ -211,26 +253,30 @@ static apr_byte_t apr_jwk_parse_rsa_x5c(apr_pool_t *pool, apr_jwk_t *jwk) {
 /*
  * parse an RSA JWK
  */
-static apr_byte_t apr_jwk_parse_rsa(apr_pool_t *pool, apr_jwk_t *jwk) {
+static apr_byte_t apr_jwk_parse_rsa(apr_pool_t *pool, json_t *json,
+		apr_jwk_t *jwk, apr_jwt_error_t *err) {
 
 	jwk->type = APR_JWK_KEY_RSA;
 
 	char *s_test = NULL;
-	apr_jwt_get_string(pool, &jwk->value, "n", &s_test);
+	apr_jwt_get_string(pool, json, "n", FALSE, &s_test, NULL);
 	if (s_test != NULL)
-		return apr_jwk_parse_rsa_raw(pool, &jwk->value, &jwk->key.rsa);
+		return apr_jwk_parse_rsa_raw(pool, json, &jwk->key.rsa, err);
 
-	json_t *v = json_object_get(jwk->value.json, "x5c");
+	json_t *v = json_object_get(json, "x5c");
 	if (v != NULL)
-		return apr_jwk_parse_rsa_x5c(pool, jwk);
+		return apr_jwk_parse_rsa_x5c(pool, json, jwk, err);
 
+	apr_jwt_error(err,
+			"wrong or unsupported RSA key representation, no \"n\" or \"x5c\" key found in JWK JSON value");
 	return FALSE;
 }
 
 /*
  * parse an EC JWK
  */
-static apr_byte_t apr_jwk_parse_ec(apr_pool_t *pool, apr_jwk_t *jwk) {
+static apr_byte_t apr_jwk_parse_ec(apr_pool_t *pool, json_t *json,
+		apr_jwk_t *jwk, apr_jwt_error_t *err) {
 
 	/* allocated space and set key type */
 	jwk->type = APR_JWK_KEY_EC;
@@ -238,65 +284,159 @@ static apr_byte_t apr_jwk_parse_ec(apr_pool_t *pool, apr_jwk_t *jwk) {
 
 	/* parse x */
 	char *s_x = NULL;
-	apr_jwt_get_string(pool, &jwk->value, "x", &s_x);
-	if (s_x == NULL)
+	if (apr_jwt_get_string(pool, json, "x", TRUE, &s_x, err) == FALSE)
 		return FALSE;
 
-	/* parse x size */
+	/* base64url decode x and get its size */
 	jwk->key.ec->x_len = apr_jwt_base64url_decode(pool,
 			(char **) &jwk->key.ec->x, s_x, 1);
+	if (jwk->key.ec->x_len <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_decode of x length failed");
+		return FALSE;
+	}
 
 	/* parse y */
 	char *s_y = NULL;
-	apr_jwt_get_string(pool, &jwk->value, "y", &s_y);
-	if (s_y == NULL)
+	if (apr_jwt_get_string(pool, json, "y", TRUE, &s_y, err) == FALSE)
 		return FALSE;
 
-	/* parse y size */
+	/* base64url decode y and get its size */
 	jwk->key.ec->y_len = apr_jwt_base64url_decode(pool,
 			(char **) &jwk->key.ec->y, s_y, 1);
+	if (jwk->key.ec->y_len <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_decode of y length failed");
+		return FALSE;
+	}
 
 	/* that went well */
 	return TRUE;
 }
 
 /*
- * parse JSON JWK
+ * parse a an octet sequence used to represent a symmetric key
  */
-apr_byte_t apr_jwk_parse_json(apr_pool_t *pool, json_t *j_json,
-		const char *s_json, apr_jwk_t **j_jwk) {
+static apr_byte_t apr_jwk_parse_oct(apr_pool_t *pool, json_t *json,
+		apr_jwk_t *jwk, apr_jwt_error_t *err) {
 
-	/* check that we've actually got a JSON value back */
-	if (j_json == NULL)
+	/* allocated space and set key type */
+	jwk->type = APR_JWK_KEY_OCT;
+	jwk->key.oct = apr_pcalloc(pool, sizeof(apr_jwk_key_oct_t));
+
+	/* parse k */
+	char *s_k = NULL;
+	if (apr_jwt_get_string(pool, json, "k", TRUE, &s_k, err) == FALSE)
 		return FALSE;
 
-	/* check that the value is a JSON object */
-	if (!json_is_object(j_json))
+	/* base64url decode k and get its size */
+	jwk->key.oct->k_len = apr_jwt_base64url_decode(pool,
+			(char **) &jwk->key.oct->k, s_k, 1);
+	if (jwk->key.oct->k_len <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_decode of k length failed");
 		return FALSE;
+	}
+
+	/* that went well */
+	return TRUE;
+}
+
+/*
+ * calculate a hash and base64url encode the result
+ */
+static apr_byte_t apr_jwk_hash_and_base64urlencode(apr_pool_t *pool,
+		const unsigned char *input, const int input_len, char **output,
+		apr_jwt_error_t *err) {
+
+	unsigned int hash_len = SHA_DIGEST_LENGTH;
+	unsigned char hash[SHA_DIGEST_LENGTH];
+
+	// TODO: upgrade to SHA2?
+
+	/* hash it */
+	if (!SHA1(input, input_len, hash)) {
+		apr_jwt_error_openssl(err, "SHA1");
+		return FALSE;
+	}
+
+	/* base64url encode the key fingerprint */
+	if (apr_jwt_base64url_encode(pool, output, (const char *) hash, hash_len, 0)
+			<= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_encode of hash failed");
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+/*
+ * parse a symmetric key in to an "oct" JWK
+ */
+apr_byte_t apr_jwk_parse_symmetric_key(apr_pool_t *pool, const unsigned char *key,
+		unsigned int key_len, apr_jwk_t **j_jwk, apr_jwt_error_t *err) {
 
 	/* allocate memory for the JWK */
 	*j_jwk = apr_pcalloc(pool, sizeof(apr_jwk_t));
 	apr_jwk_t *jwk = *j_jwk;
 
-	/* set the raw JSON/string representations */
-	jwk->value.json = j_json;
-	jwk->value.str = apr_pstrdup(pool, s_json);
+	/* allocated space and set key type */
+	jwk->type = APR_JWK_KEY_OCT;
+	jwk->key.oct = apr_pcalloc(pool, sizeof(apr_jwk_key_oct_t));
 
-	/* get the key type */
+	//	/* set the values */
+	jwk->key.oct->k = apr_pcalloc(pool, key_len);
+	memcpy(jwk->key.oct->k, key, key_len);
+	jwk->key.oct->k_len = key_len;
+
+	/* calculate a unique key identifier (kid) by fingerprinting the key params */
+	if (apr_jwk_hash_and_base64urlencode(pool, jwk->key.oct->k,
+			jwk->key.oct->k_len, &jwk->kid, err) == FALSE)
+		return FALSE;
+
+	return TRUE;
+}
+
+/*
+ * parse JSON JWK
+ */
+apr_byte_t apr_jwk_parse_json(apr_pool_t *pool, json_t *json, apr_jwk_t **j_jwk,
+		apr_jwt_error_t *err) {
+
+	/* check that we've actually got a JSON value back */
+	if (json == NULL) {
+		apr_jwt_error(err, "JWK JSON is NULL");
+		return FALSE;
+	}
+
+	/* check that the value is a JSON object */
+	if (!json_is_object(json)) {
+		apr_jwt_error(err, "JWK JSON is not a JSON object");
+		return FALSE;
+	}
+
+	/* allocate memory for the JWK */
+	*j_jwk = apr_pcalloc(pool, sizeof(apr_jwk_t));
+	apr_jwk_t *jwk = *j_jwk;
+
+	/* get the mandatory key type */
 	char *kty = NULL;
-	if (apr_jwt_get_string(pool, &jwk->value, "kty", &kty) == FALSE)
+	if (apr_jwt_get_string(pool, json, "kty", TRUE, &kty, err) == FALSE)
 		return FALSE;
 
-	/* kty is mandatory */
-	if (kty == NULL)
-		return FALSE;
+	/* get the optional kid */
+	apr_jwt_get_string(pool, json, "kid", FALSE, &jwk->kid, NULL);
 
 	/* parse the key */
 	if (apr_strnatcmp(kty, "RSA") == 0)
-		return apr_jwk_parse_rsa(pool, jwk);
+		return apr_jwk_parse_rsa(pool, json, jwk, err);
 
 	if (apr_strnatcmp(kty, "EC") == 0)
-		return apr_jwk_parse_ec(pool, jwk);
+		return apr_jwk_parse_ec(pool, json, jwk, err);
+
+	if (apr_strnatcmp(kty, "oct") == 0)
+		return apr_jwk_parse_oct(pool, json, jwk, err);
+
+	apr_jwt_error(err,
+			"wrong or unsupported JWK key representation \"%s\" (\"RSA\", \"EC\" and \"oct\" are supported key types)",
+			kty);
 
 	return FALSE;
 }
@@ -304,67 +444,89 @@ apr_byte_t apr_jwk_parse_json(apr_pool_t *pool, json_t *j_json,
 /*
  * convert RSA key to JWK JSON string representation and kid
  */
-static apr_byte_t apr_jwk_rsa_to_json(apr_pool_t *pool, apr_jwk_key_rsa_t *key,
-		char **jwk, char**kid) {
+apr_byte_t apr_jwk_to_json(apr_pool_t *pool, apr_jwk_t *jwk, char **s_json,
+		apr_jwt_error_t *err) {
+
+	if (jwk->type != APR_JWK_KEY_RSA) {
+		apr_jwt_error(err, "non RSA keys (%d) not yet supported", jwk->type);
+		return FALSE;
+	}
+
+	apr_jwk_key_rsa_t *key = jwk->key.rsa;
 
 	unsigned char *n_enc = NULL;
 	int n_len = apr_jwt_base64url_encode(pool, (char **) &n_enc,
 			(const char *) key->modulus, key->modulus_len, 0);
+	if (n_len <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_encode of modulus failed");
+		return FALSE;
+	}
 
 	unsigned char *e_enc = NULL;
-	apr_jwt_base64url_encode(pool, (char **) &e_enc,
-			(const char *) key->exponent, key->exponent_len, 0);
+	if (apr_jwt_base64url_encode(pool, (char **) &e_enc,
+			(const char *) key->exponent, key->exponent_len, 0) <= 0) {
+		apr_jwt_error(err,
+				"apr_jwt_base64url_encode of public exponent failed");
+		return FALSE;
+	}
 
 	unsigned char *d_enc = NULL;
-	if (key->private_exponent_len > 0)
-		apr_jwt_base64url_encode(pool, (char **) &d_enc,
+	if (key->private_exponent_len > 0) {
+		if (apr_jwt_base64url_encode(pool, (char **) &d_enc,
 				(const char *) key->private_exponent, key->private_exponent_len,
-				0);
-
-	/* calculate a unique key identifier (kid) by fingerprinting the key params */
-	// TODO: based just on sha1 hash of baseurl-encoded "n" now...
-	unsigned int fp_len = SHA_DIGEST_LENGTH;
-	unsigned char fp[SHA_DIGEST_LENGTH];
-	if (!SHA1(n_enc, n_len, fp))
-		return FALSE;
-	char *fp_enc = NULL;
-	if (apr_jwt_base64url_encode(pool, &fp_enc, (const char *) fp, fp_len,
-			0) == FALSE)
-		return FALSE;
+				0) <= 0) {
+			apr_jwt_error(err,
+					"apr_jwt_base64url_encode of private exponent failed");
+			return FALSE;
+		}
+	}
 
 	char *p = apr_psprintf(pool, "{ \"kty\" : \"RSA\"");
 	p = apr_psprintf(pool, "%s, \"n\": \"%s\"", p, n_enc);
 	p = apr_psprintf(pool, "%s, \"e\": \"%s\"", p, e_enc);
 	if (d_enc != NULL)
 		p = apr_psprintf(pool, "%s, \"d\": \"%s\"", p, d_enc);
-	p = apr_psprintf(pool, "%s, \"kid\" : \"%s\"", p, fp_enc);
+	p = apr_psprintf(pool, "%s, \"kid\" : \"%s\"", p, jwk->kid);
 	p = apr_psprintf(pool, "%s }", p);
 
-	*jwk = p;
-	*kid = fp_enc;
+	*s_json = p;
 
 	return TRUE;
 }
 
-/*
- * convert PEM formatted public/private key file to JSON string representation
- */
-static apr_byte_t apr_jwk_pem_to_json_impl(apr_pool_t *pool,
-		const char *filename, char **s_jwk, char**s_kid, int is_private_key) {
+static apr_byte_t apr_jwk_parse_rsa_key(apr_pool_t *pool, int is_private_key,
+		const char *filename, apr_jwk_t **j_jwk, apr_jwt_error_t *err) {
 	BIO *input = NULL;
 	apr_jwk_key_rsa_t *key = NULL;
 	apr_byte_t rv = FALSE;
 
-	if ((input = BIO_new(BIO_s_file())) == NULL)
+	if ((input = BIO_new(BIO_s_file())) == NULL) {
+		apr_jwt_error_openssl(err, "BIO_new/BIO_s_file");
+		goto end;
+	}
+
+	if (BIO_read_filename(input, filename) <= 0) {
+		apr_jwt_error_openssl(err, "BIO_read_filename");
+		goto end;
+	}
+
+	if (apr_jwk_rsa_bio_to_key(pool, input, &key, is_private_key, err) == FALSE)
 		goto end;
 
-	if (BIO_read_filename(input, filename) <= 0)
+	/* allocate memory for the JWK */
+	*j_jwk = apr_pcalloc(pool, sizeof(apr_jwk_t));
+	apr_jwk_t *jwk = *j_jwk;
+
+	jwk->type = APR_JWK_KEY_RSA;
+	jwk->key.rsa = key;
+
+	/* calculate a unique key identifier (kid) by fingerprinting the key params */
+	// TODO: based just on sha1 hash of modulus "n" now..., could do this based on jwk->value.str
+	if (apr_jwk_hash_and_base64urlencode(pool, key->modulus, key->modulus_len,
+			&jwk->kid, err) == FALSE)
 		goto end;
 
-	if (apr_jwk_rsa_bio_to_key(pool, input, &key, is_private_key) == FALSE)
-		goto end;
-
-	rv = apr_jwk_rsa_to_json(pool, key, s_jwk, s_kid);
+	rv = TRUE;
 
 end:
 
@@ -374,20 +536,12 @@ end:
 	return rv;
 }
 
-/*
- * convert the RSA public key in the X.509 certificate in the file pointed to
- * by "filename" to a JSON Web Key string representation
- */
-apr_byte_t apr_jwk_pem_to_json(apr_pool_t *pool, const char *filename,
-		char **s_jwk, char**s_kid) {
-	return apr_jwk_pem_to_json_impl(pool, filename, s_jwk, s_kid, FALSE);
+apr_byte_t apr_jwk_parse_rsa_private_key(apr_pool_t *pool, const char *filename,
+		apr_jwk_t **j_jwk, apr_jwt_error_t *err) {
+	return apr_jwk_parse_rsa_key(pool, TRUE, filename, j_jwk, err);
 }
 
-/*
- * convert the RSA private key in the PEM file pointed to by "filename"
- * to a JSON Web Key string representation
- */
-apr_byte_t apr_jwk_private_key_to_rsa_jwk(apr_pool_t *pool,
-		const char *filename, char **s_jwk, char**s_kid) {
-	return apr_jwk_pem_to_json_impl(pool, filename, s_jwk, s_kid, TRUE);
+apr_byte_t apr_jwk_parse_rsa_public_key(apr_pool_t *pool, const char *filename,
+		apr_jwk_t **j_jwk, apr_jwt_error_t *err) {
+	return apr_jwk_parse_rsa_key(pool, FALSE, filename, j_jwk, err);
 }
