@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -53,6 +53,27 @@
 #include <apr_base64.h>
 
 #include "apr_jose.h"
+#include <openssl/opensslv.h>
+
+#ifdef WIN32
+#define snprintf _snprintf
+#endif
+
+/*
+ * assemble an error report
+ */
+void _apr_jwt_error_set(apr_jwt_error_t *error, const char *source,
+		const int line, const char *function, const char *msg, ...) {
+	if (error == NULL)
+		return;
+	snprintf(error->source, APR_JWT_ERROR_SOURCE_LENGTH, "%s", source);
+	error->line = line;
+	snprintf(error->function, APR_JWT_ERROR_FUNCTION_LENGTH, "%s", function);
+	va_list ap;
+	va_start(ap, msg);
+	vsnprintf(error->text, APR_JWT_ERROR_TEXT_LENGTH, msg, ap);
+	va_end(ap);
+}
 
 /*
  * check if a string is an element of an array of strings
@@ -138,26 +159,30 @@ int apr_jwt_base64url_decode(apr_pool_t *pool, char **dst, const char *src,
  * parse JSON object from string in to JWT value
  */
 static apr_byte_t apr_jwt_base64url_decode_object(apr_pool_t *pool,
-		const char *str, apr_jwt_value_t *value) {
+		const char *str, apr_jwt_value_t *value, apr_jwt_error_t *err) {
 
 	/* base64url-decode the string representation into value->str */
-	if (apr_jwt_base64url_decode(pool, &value->str, str, 1) < 0)
+	if (apr_jwt_base64url_decode(pool, &value->str, str, 1) <= 0) {
+		apr_jwt_error(err, "apr_jwt_base64url_decode of (%s) failed", str);
 		return FALSE;
+	}
 
 	/* decode the string in to a JSON structure into value->json */
 	json_error_t json_error;
 	value->json = json_loads(value->str, 0, &json_error);
 
-	if (value->json == NULL)
-		return FALSE;
-
 	/* check that we've actually got a JSON value back */
-	if (value->json == NULL)
+	if (value->json == NULL) {
+		apr_jwt_error(err, "JSON parsing (json_loads) failed: %s (%s)",
+				json_error.text, value->str);
 		return FALSE;
+	}
 
 	/* check that the value is a JSON object */
-	if (!json_is_object(value->json))
+	if (!json_is_object(value->json)) {
+		apr_jwt_error(err, "JSON value is not an object");
 		return FALSE;
+	}
 
 	return TRUE;
 }
@@ -165,13 +190,23 @@ static apr_byte_t apr_jwt_base64url_decode_object(apr_pool_t *pool,
 /*
  * get (optional) string from JWT
  */
-apr_byte_t apr_jwt_get_string(apr_pool_t *pool, apr_jwt_value_t *value,
-		const char *claim_name, char **result) {
-	json_t *v = json_object_get(value->json, claim_name);
-	if ((v != NULL) && (json_is_string(v))) {
-		*result = apr_pstrdup(pool, json_string_value(v));
-	} else {
-		*result = NULL;
+apr_byte_t apr_jwt_get_string(apr_pool_t *pool, json_t *json,
+		const char *claim_name, apr_byte_t is_mandatory, char **result,
+		apr_jwt_error_t *err) {
+	json_t *v = json_object_get(json, claim_name);
+	if (v != NULL) {
+		if (json_is_string(v)) {
+			*result = apr_pstrdup(pool, json_string_value(v));
+		} else if (is_mandatory) {
+			apr_jwt_error(err,
+					"mandatory JSON key \"%s\" was found but the type is not a string",
+					claim_name);
+			return FALSE;
+		}
+	} else if (is_mandatory) {
+		apr_jwt_error(err, "mandatory JSON key \"%s\" could not be found",
+				claim_name);
+		return FALSE;
 	}
 	return TRUE;
 }
@@ -179,13 +214,24 @@ apr_byte_t apr_jwt_get_string(apr_pool_t *pool, apr_jwt_value_t *value,
 /*
  * parse (optional) timestamp from payload
  */
-static apr_byte_t apr_jwt_parse_timestamp(apr_pool_t *pool,
-		apr_jwt_value_t *value, const char *claim_name, apr_time_t *result) {
-	json_t *v = json_object_get(value->json, claim_name);
-	if ((v != NULL) && (json_is_integer(v))) {
-		*result = apr_time_from_sec(json_integer_value(v));
-	} else {
-		*result = APR_JWT_CLAIM_TIME_EMPTY;
+static apr_byte_t apr_jwt_get_timestamp(apr_pool_t *pool, json_t *json,
+		const char *claim_name, apr_byte_t is_mandatory, json_int_t *result,
+		apr_jwt_error_t *err) {
+	*result = APR_JWT_CLAIM_TIME_EMPTY;
+	json_t *v = json_object_get(json, claim_name);
+	if (v != NULL) {
+		if (json_is_integer(v)) {
+			*result = json_integer_value(v);
+		} else if (is_mandatory) {
+			apr_jwt_error(err,
+					"mandatory JSON key \"%s\" was found but the type is not a number",
+					claim_name);
+			return FALSE;
+		}
+	} else if (is_mandatory) {
+		apr_jwt_error(err, "mandatory JSON key \"%s\" could not be found",
+				claim_name);
+		return FALSE;
 	}
 	return TRUE;
 }
@@ -193,25 +239,26 @@ static apr_byte_t apr_jwt_parse_timestamp(apr_pool_t *pool,
 /*
  * parse a JWT header
  */
-apr_byte_t apr_jwt_parse_header(apr_pool_t *pool, const char *s_header,
-		apr_jwt_header_t *header) {
+static apr_byte_t apr_jwt_parse_header_object(apr_pool_t *pool,
+		const char *s_header, apr_jwt_header_t *header, apr_jwt_error_t *err) {
 
 	/* decode the JWT JSON header */
-	if (apr_jwt_base64url_decode_object(pool, s_header, &header->value) == FALSE)
+	if (apr_jwt_base64url_decode_object(pool, s_header, &header->value,
+			err) == FALSE)
 		return FALSE;
 
-	/* parse the (optional) signing algorithm */
-	apr_jwt_get_string(pool, &header->value, "alg", &header->alg);
-
-	/* check that the mandatory algorithm was set */
-	if (header->alg == NULL)
+	/* parse the (mandatory) signing algorithm */
+	if (apr_jwt_get_string(pool, header->value.json, "alg", TRUE, &header->alg,
+			err) == FALSE)
 		return FALSE;
 
 	/* parse the (optional) kid */
-	apr_jwt_get_string(pool, &header->value, "kid", &header->kid);
+	apr_jwt_get_string(pool, header->value.json, "kid", FALSE, &header->kid,
+			NULL);
 
 	/* parse the (optional) enc */
-	apr_jwt_get_string(pool, &header->value, "enc", &header->enc);
+	apr_jwt_get_string(pool, header->value.json, "enc", FALSE, &header->enc,
+			NULL);
 
 	return TRUE;
 }
@@ -220,24 +267,30 @@ apr_byte_t apr_jwt_parse_header(apr_pool_t *pool, const char *s_header,
  * parse JWT payload
  */
 static apr_byte_t apr_jwt_parse_payload(apr_pool_t *pool, const char *s_payload,
-		apr_jwt_payload_t *payload) {
+		apr_jwt_payload_t *payload, apr_jwt_error_t *err) {
 
 	/* decode the JWT JSON payload */
-	if (apr_jwt_base64url_decode_object(pool, s_payload,
-			&payload->value) == FALSE)
+	if (apr_jwt_base64url_decode_object(pool, s_payload, &payload->value,
+			err) == FALSE)
 		return FALSE;
 
 	/* get the (optional) "issuer" value from the JSON payload */
-	apr_jwt_get_string(pool, &payload->value, "iss", &payload->iss);
+	apr_jwt_get_string(pool, payload->value.json, "iss", FALSE, &payload->iss,
+			NULL);
 
 	/* get the (optional) "exp" value from the JSON payload */
-	apr_jwt_parse_timestamp(pool, &payload->value, "exp", &payload->exp);
+	apr_jwt_get_timestamp(pool, payload->value.json, "exp", FALSE,
+			&payload->exp,
+			NULL);
 
 	/* get the (optional) "iat" value from the JSON payload */
-	apr_jwt_parse_timestamp(pool, &payload->value, "iat", &payload->iat);
+	apr_jwt_get_timestamp(pool, payload->value.json, "iat", FALSE,
+			&payload->iat,
+			NULL);
 
 	/* get the (optional) "sub" value from the JSON payload */
-	apr_jwt_get_string(pool, &payload->value, "sub", &payload->sub);
+	apr_jwt_get_string(pool, payload->value.json, "sub", FALSE, &payload->sub,
+			NULL);
 
 	return TRUE;
 }
@@ -257,7 +310,7 @@ static apr_byte_t apr_jwt_parse_signature(apr_pool_t *pool,
 /*
  * parse a JWT that uses compact serialization (i.e. elements separated by dots) in to an array of strings
  */
-apr_array_header_t *apr_jwt_compact_deserialize(apr_pool_t *pool,
+static apr_array_header_t *apr_jwt_compact_deserialize(apr_pool_t *pool,
 		const char *str) {
 	apr_array_header_t *result = apr_array_make(pool, 6, sizeof(const char*));
 	if ((str != NULL) && (strlen(str) > 0)) {
@@ -275,53 +328,116 @@ apr_array_header_t *apr_jwt_compact_deserialize(apr_pool_t *pool,
 	return result;
 }
 
+/*
+ * parse a JOSE header from a compact serialized string
+ */
+apr_byte_t apr_jwt_header_parse(apr_pool_t *pool, const char *s_json,
+		apr_array_header_t **unpacked, apr_jwt_header_t *header,
+		apr_jwt_error_t *err) {
+	*unpacked = apr_jwt_compact_deserialize(pool, s_json);
+	if ((*unpacked)->nelts < 1) {
+		apr_jwt_error(err, "could not deserialize at least one element");
+		return FALSE;
+	}
+	if (apr_jwt_parse_header_object(pool, ((const char**) (*unpacked)->elts)[0],
+			header, err) == FALSE)
+		return FALSE;
+	return TRUE;
+}
+
 /*
  * return plain deserialized header text
  */
-const char *apr_jwt_header_to_string(apr_pool_t *pool, const char *s_json) {
-	apr_array_header_t *unpacked = apr_jwt_compact_deserialize(pool, s_json);
-	if (unpacked->nelts < 1)
-		return NULL;
+const char *apr_jwt_header_to_string(apr_pool_t *pool, const char *s_json,
+		apr_jwt_error_t *err) {
+	apr_array_header_t *unpacked = NULL;
 	apr_jwt_header_t header;
-	if (apr_jwt_parse_header(pool, ((const char**) unpacked->elts)[0],
-			&header) == FALSE)
+	if (apr_jwt_header_parse(pool, s_json, &unpacked, &header, err) == FALSE)
 		return NULL;
 	json_decref(header.value.json);
 	return header.value.str;
 }
 
 /*
- * parse a JSON Web Token
+ * return the JWK type for the JWT signature verification
+ */
+const char *apr_jwt_signature_to_jwk_type(apr_pool_t *pool, apr_jwt_t *jwt) {
+	if (apr_jws_signature_is_rsa(pool, jwt)) {
+		return "RSA";
+	}
+#if (OPENSSL_VERSION_NUMBER >= 0x01000000)
+	if (apr_jws_signature_is_ec(pool, jwt)) {
+		return "EC";
+	}
+#endif
+	if (apr_jws_signature_is_hmac(pool, jwt)) {
+		return "oct";
+	}
+	return NULL;
+}
+
+/* see if we can deal with this type of JWT (JWS/JWE) */
+static apr_byte_t apr_jwt_is_supported(apr_pool_t *pool, apr_jwt_t *jwt,
+		apr_jwt_error_t *err) {
+	if (apr_jws_algorithm_is_supported(pool, jwt->header.alg) == FALSE) {
+		if (apr_jwe_algorithm_is_supported(pool, jwt->header.alg) == FALSE) {
+			apr_jwt_error(err, "unsupported algorithm in JWT header: \"%s\"",
+					jwt->header.alg);
+			return FALSE;
+		}
+		if (apr_jwe_encryption_is_supported(pool, jwt->header.enc) == FALSE) {
+			apr_jwt_error(err,
+					"unsupported content encryption algorithm in (%s) encrypted JWT header: \"%s\"",
+					jwt->header.alg, jwt->header.enc);
+			return FALSE;
+		}
+	}
+	return TRUE;
+}
+
+/*
+ * parse and (optionally) decrypt a JSON Web Token
  */
 apr_byte_t apr_jwt_parse(apr_pool_t *pool, const char *s_json,
-		apr_jwt_t **j_jwt, apr_hash_t *private_keys, const char *shared_key) {
+		apr_jwt_t **j_jwt, apr_hash_t *keys, apr_jwt_error_t *err) {
 
 	*j_jwt = apr_pcalloc(pool, sizeof(apr_jwt_t));
 	apr_jwt_t *jwt = *j_jwt;
 
-	apr_array_header_t *unpacked = apr_jwt_compact_deserialize(pool, s_json);
-	if (unpacked->nelts < 2)
+	apr_array_header_t *unpacked = NULL;
+	if (apr_jwt_header_parse(pool, s_json, &unpacked, &jwt->header,
+			err) == FALSE)
 		return FALSE;
 
-	/* parse the header fields */
-	if (apr_jwt_parse_header(pool, ((const char**) unpacked->elts)[0],
-			&jwt->header) == FALSE)
+	if (unpacked->nelts < 2) {
+		apr_jwt_error(err,
+				"could not successfully deserialize 2 or more elements from JWT header");
+		return FALSE;
+	}
+
+	if (apr_jwt_is_supported(pool, jwt, err) == FALSE)
 		return FALSE;
 
 	if (apr_jwe_is_encrypted_jwt(pool, &jwt->header)) {
+
 		char *decrypted = NULL;
-		if ((apr_jwe_decrypt_jwt(pool, &jwt->header, unpacked, private_keys,
-				shared_key, &decrypted) == FALSE) || (decrypted == NULL))
+		if ((apr_jwe_decrypt_jwt(pool, &jwt->header, unpacked, keys, &decrypted,
+				err) == FALSE) || (decrypted == NULL))
 			return FALSE;
+
 		apr_array_clear(unpacked);
-		unpacked = apr_jwt_compact_deserialize(pool, (const char *) decrypted);
+		unpacked = NULL;
 		json_decref(jwt->header.value.json);
-		if (unpacked->nelts < 2)
+
+		if (apr_jwt_header_parse(pool, (const char *) decrypted, &unpacked,
+				&jwt->header, err) == FALSE)
 			return FALSE;
-		/* parse the nested header fields */
-		if (apr_jwt_parse_header(pool, ((const char**) unpacked->elts)[0],
-				&jwt->header) == FALSE)
+
+		if (unpacked->nelts < 2) {
+			apr_jwt_error(err,
+					"could not successfully deserialize 2 or more elements from decrypted JWT header");
 			return FALSE;
+		}
 	}
 
 	/* concat the base64url-encoded payload to the base64url-encoded header for signature verification purposes */
@@ -330,7 +446,7 @@ apr_byte_t apr_jwt_parse(apr_pool_t *pool, const char *s_json,
 
 	/* parse the payload fields */
 	if (apr_jwt_parse_payload(pool, ((const char**) unpacked->elts)[1],
-			&jwt->payload) == FALSE) {
+			&jwt->payload, err) == FALSE) {
 		json_decref(jwt->header.value.json);
 		return FALSE;
 	}
@@ -341,6 +457,8 @@ apr_byte_t apr_jwt_parse(apr_pool_t *pool, const char *s_json,
 				&jwt->signature) == FALSE) {
 			json_decref(jwt->header.value.json);
 			json_decref(jwt->payload.value.json);
+			apr_jwt_error(err,
+					"could not successfully parse (base64urldecode) JWT signature");
 			return FALSE;
 		}
 	}
@@ -357,3 +475,18 @@ void apr_jwt_destroy(apr_jwt_t *jwt) {
 			json_decref(jwt->payload.value.json);
 	}
 }
+
+/* timing-safe byte sequence comparison */
+apr_byte_t apr_jwt_memcmp(const void *in_a, const void *in_b, size_t len) {
+	// TODO: this is copied from OpenSSL 1.0.1 to avoid version issues on various platforms
+	//       we could use #ifdef's to use OpenSSL where possible
+	size_t i;
+	const unsigned char *a = in_a;
+	const unsigned char *b = in_b;
+	unsigned char x = 0;
+
+	for (i = 0; i < len; i++)
+		x |= a[i] ^ b[i];
+
+	return x ? FALSE : TRUE;
+}
