@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -102,8 +102,7 @@ static const char *oidc_metadata_filename_to_issuer(request_rec *r,
 	char *p = strrchr(result, '.');
 	*p = '\0';
 	p = oidc_util_unescape_string(r, result);
-	return (strcmp(p, "accounts.google.com") == 0) ?
-			p : apr_psprintf(r->pool, "https://%s", p);
+	return apr_psprintf(r->pool, "https://%s", p);
 }
 
 /*
@@ -146,11 +145,11 @@ static const char *oidc_metadata_conf_path(request_rec *r, const char *issuer) {
 }
 
 /*
- * get the full path to the jwks metadata file for a specified issuer
+ * get cache key for the JWKs file for a specified URI
  */
 static const char *oidc_metadata_jwks_cache_key(request_rec *r,
-		const char *issuer) {
-	return apr_psprintf(r->pool, "%s.jwks", issuer);
+		const char *jwks_uri) {
+	return jwks_uri;
 }
 
 /*
@@ -425,14 +424,14 @@ static apr_byte_t oidc_metadata_client_is_valid(request_rec *r,
 /*
  * checks if a parsed JWKs file is a valid one, cq. contains "keys"
  */
-static apr_byte_t oidc_metadata_jwks_is_valid(request_rec *r, json_t *j_jwks,
-		const char *issuer) {
+static apr_byte_t oidc_metadata_jwks_is_valid(request_rec *r,
+		const oidc_jwks_uri_t *jwks_uri, json_t *j_jwks) {
 
 	json_t *keys = json_object_get(j_jwks, "keys");
 	if ((keys == NULL) || (!json_is_array(keys))) {
 		oidc_error(r,
-				"provider (%s) JWKS JSON metadata did not contain a \"keys\" array",
-				issuer);
+				"JWKs JSON metadata obtained from URL \"%s\" did not contain a \"keys\" array",
+				jwks_uri->url);
 		return FALSE;
 	}
 	return TRUE;
@@ -561,6 +560,10 @@ static apr_byte_t oidc_metadata_file_write(request_rec *r, const char *path,
 static apr_byte_t oidc_metadata_client_register(request_rec *r, oidc_cfg *cfg,
 		oidc_provider_t *provider, json_t **j_client, const char **response) {
 
+	/* get a handle to the directory config */
+	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+			&auth_openidc_module);
+
 	/* assemble the JSON registration request */
 	json_t *data = json_object();
 	json_object_set_new(data, "client_name",
@@ -621,45 +624,95 @@ static apr_byte_t oidc_metadata_client_register(request_rec *r, oidc_cfg *cfg,
 	json_object_set_new(data, "initiate_login_uri",
 			json_string(cfg->redirect_uri));
 
+	json_object_set_new(data, "logout_uri",
+			json_string(
+					apr_psprintf(r->pool, "%s?logout=%s", cfg->redirect_uri,
+							OIDC_GET_STYLE_LOGOUT_PARAM_VALUE)));
+
+	if (cfg->default_slo_url != NULL) {
+		json_object_set_new(data, "post_logout_redirect_uris",
+				json_pack("[s]", cfg->default_slo_url));
+	}
+
+	if (provider->registration_endpoint_json != NULL) {
+
+		json_error_t json_error;
+		json_t *json = json_loads(provider->registration_endpoint_json, 0,
+				&json_error);
+
+		if (json == NULL) {
+
+			oidc_error(r, "JSON parsing returned an error: %s",
+					json_error.text);
+
+		} else {
+
+			if (!json_is_object(json)) {
+
+				oidc_error(r, "parsed JSON did not contain a JSON object");
+
+			} else {
+
+				oidc_util_json_merge(json, data);
+
+			}
+
+			json_decref(json);
+		}
+	}
+
 	/* dynamically register the client with the specified parameters */
 	if (oidc_util_http_post_json(r, provider->registration_endpoint_url, data,
 			NULL, provider->registration_token, provider->ssl_validate_server, response,
-			cfg->http_timeout_short, cfg->outgoing_proxy) == FALSE) {
+			cfg->http_timeout_short, cfg->outgoing_proxy,
+			dir_cfg->pass_cookies) == FALSE) {
 		json_decref(data);
 		return FALSE;
 	}
 	json_decref(data);
 
 	/* decode and see if it is not an error response somehow */
-	return oidc_util_decode_json_and_check_error(r, *response, j_client);
+	if (oidc_util_decode_json_and_check_error(r, *response, j_client) == FALSE) {
+		oidc_error(r,
+				"JSON parsing of dynamic client registration response failed");
+		return FALSE;
+	}
+
+	return TRUE;
 }
 
 /*
  * helper function to get the JWKs for the specified issuer
  */
 static apr_byte_t oidc_metadata_jwks_retrieve_and_cache(request_rec *r,
-		oidc_cfg *cfg, oidc_provider_t *provider, json_t **j_jwks) {
+		oidc_cfg *cfg, const oidc_jwks_uri_t *jwks_uri, json_t **j_jwks) {
 
 	const char *response = NULL;
 
+	/* get a handle to the directory config */
+	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+			&auth_openidc_module);
+
 	/* no valid provider metadata, get it at the specified URL with the specified parameters */
-	if (oidc_util_http_get(r, provider->jwks_uri, NULL, NULL,
-			NULL, provider->ssl_validate_server, &response, cfg->http_timeout_long,
-			cfg->outgoing_proxy) == FALSE)
+	if (oidc_util_http_get(r, jwks_uri->url, NULL, NULL,
+			NULL, jwks_uri->ssl_validate_server, &response, cfg->http_timeout_long,
+			cfg->outgoing_proxy, dir_cfg->pass_cookies) == FALSE)
 		return FALSE;
 
 	/* decode and see if it is not an error response somehow */
-	if (oidc_util_decode_json_and_check_error(r, response, j_jwks) == FALSE)
+	if (oidc_util_decode_json_and_check_error(r, response, j_jwks) == FALSE) {
+		oidc_error(r, "JSON parsing of JWKs published at the jwks_uri failed");
 		return FALSE;
+	}
 
 	/* check to see if it is valid metadata */
-	if (oidc_metadata_jwks_is_valid(r, *j_jwks, provider->issuer) == FALSE)
+	if (oidc_metadata_jwks_is_valid(r, jwks_uri, *j_jwks) == FALSE)
 		return FALSE;
 
 	/* store the JWKs in the cache */
 	cfg->cache->set(r, OIDC_CACHE_SECTION_JWKS,
-			oidc_metadata_jwks_cache_key(r, provider->issuer), response,
-			apr_time_now() + apr_time_from_sec(provider->jwks_refresh_interval));
+			oidc_metadata_jwks_cache_key(r, jwks_uri->url), response,
+			apr_time_now() + apr_time_from_sec(jwks_uri->refresh_interval));
 
 	return TRUE;
 }
@@ -668,15 +721,15 @@ static apr_byte_t oidc_metadata_jwks_retrieve_and_cache(request_rec *r,
  * return JWKs for the specified issuer
  */
 apr_byte_t oidc_metadata_jwks_get(request_rec *r, oidc_cfg *cfg,
-		oidc_provider_t *provider, json_t **j_jwks, apr_byte_t *refresh) {
+		const oidc_jwks_uri_t *jwks_uri, json_t **j_jwks, apr_byte_t *refresh) {
 
-	oidc_debug(r, "enter, issuer=%s, refresh=%d", provider->issuer, *refresh);
+	oidc_debug(r, "enter, jwks_uri=%s, refresh=%d", jwks_uri->url, *refresh);
 
 	/* see if we need to do a forced refresh */
 	if (*refresh == TRUE) {
-		oidc_debug(r, "doing a forced refresh of the JWKs for issuer \"%s\"",
-				provider->issuer);
-		if (oidc_metadata_jwks_retrieve_and_cache(r, cfg, provider,
+		oidc_debug(r, "doing a forced refresh of the JWKs from URI \"%s\"",
+				jwks_uri->url);
+		if (oidc_metadata_jwks_retrieve_and_cache(r, cfg, jwks_uri,
 				j_jwks) == TRUE)
 			return TRUE;
 		// else: fallback on any cached JWKs
@@ -685,17 +738,19 @@ apr_byte_t oidc_metadata_jwks_get(request_rec *r, oidc_cfg *cfg,
 	/* see if the JWKs is cached */
 	const char *value = NULL;
 	cfg->cache->get(r, OIDC_CACHE_SECTION_JWKS,
-			oidc_metadata_jwks_cache_key(r, provider->issuer), &value);
+			oidc_metadata_jwks_cache_key(r, jwks_uri->url), &value);
 
 	if (value == NULL) {
 		/* it is non-existing or expired: do a forced refresh */
 		*refresh = TRUE;
-		return oidc_metadata_jwks_retrieve_and_cache(r, cfg, provider, j_jwks);
+		return oidc_metadata_jwks_retrieve_and_cache(r, cfg, jwks_uri, j_jwks);
 	}
 
 	/* decode and see if it is not an error response somehow */
-	if (oidc_util_decode_json_and_check_error(r, value, j_jwks) == FALSE)
+	if (oidc_util_decode_json_and_check_error(r, value, j_jwks) == FALSE) {
+		oidc_error(r, "JSON parsing of cached JWKs data failed");
 		return FALSE;
+	}
 
 	return TRUE;
 }
@@ -707,15 +762,22 @@ apr_byte_t oidc_metadata_provider_retrieve(request_rec *r, oidc_cfg *cfg,
 		const char *issuer, const char *url, json_t **j_metadata,
 		const char **response) {
 
+	/* get a handle to the directory config */
+	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+			&auth_openidc_module);
+
 	/* get provider metadata from the specified URL with the specified parameters */
 	if (oidc_util_http_get(r, url, NULL, NULL, NULL,
 			cfg->provider.ssl_validate_server, response,
-			cfg->http_timeout_short, cfg->outgoing_proxy) == FALSE)
+			cfg->http_timeout_short, cfg->outgoing_proxy,
+			dir_cfg->pass_cookies) == FALSE)
 		return FALSE;
 
 	/* decode and see if it is not an error response somehow */
-	if (oidc_util_decode_json_and_check_error(r, *response, j_metadata) == FALSE)
+	if (oidc_util_decode_json_and_check_error(r, *response, j_metadata) == FALSE) {
+		oidc_error(r, "JSON parsing of retrieved Discovery document failed");
 		return FALSE;
+	}
 
 	/* check to see if it is valid metadata */
 	if (oidc_metadata_provider_is_valid(r, *j_metadata, issuer) == FALSE)
@@ -730,7 +792,7 @@ apr_byte_t oidc_metadata_provider_retrieve(request_rec *r, oidc_cfg *cfg,
  * if not, use OpenID Connect Discovery to get it, check it and store it
  */
 static apr_byte_t oidc_metadata_provider_get(request_rec *r, oidc_cfg *cfg,
-		const char *issuer, json_t **j_provider) {
+		const char *issuer, json_t **j_provider, apr_byte_t allow_discovery) {
 
 	/* holds the response data/string/JSON from the OP */
 	const char *response = NULL;
@@ -745,6 +807,13 @@ static apr_byte_t oidc_metadata_provider_get(request_rec *r, oidc_cfg *cfg,
 		return oidc_metadata_provider_is_valid(r, *j_provider, issuer);
 	}
 
+	if (!allow_discovery) {
+		oidc_warn(r,
+				"no metadata found for the requested issuer (%s), and Discovery is not allowed",
+				issuer);
+		return FALSE;
+	}
+
 	// TODO: how to do validity/expiry checks on provider metadata
 
 	/* assemble the URL to the .well-known OpenID metadata */
@@ -874,7 +943,7 @@ apr_byte_t oidc_metadata_list(request_rec *r, oidc_cfg *cfg,
 
 		/* get the provider and client metadata, do all checks and registration if possible */
 		oidc_provider_t *provider = NULL;
-		if (oidc_metadata_get(r, cfg, issuer, &provider) == TRUE) {
+		if (oidc_metadata_get(r, cfg, issuer, &provider, FALSE) == TRUE) {
 			/* push the decoded issuer filename in to the array */
 			*(const char**) apr_array_push(*list) = provider->issuer;
 		}
@@ -1040,6 +1109,11 @@ apr_byte_t oidc_metadata_conf_parse(request_rec *r, oidc_cfg *cfg,
 	oidc_json_object_get_int(r->pool, j_conf, "idtoken_iat_slack",
 			&provider->idtoken_iat_slack, cfg->provider.idtoken_iat_slack);
 
+	/* see if we've got a custom max session duration */
+	oidc_json_object_get_int(r->pool, j_conf, "session_max_duration",
+			&provider->session_max_duration,
+			cfg->provider.session_max_duration);
+
 	/* see if we've got custom authentication request parameter values */
 	oidc_json_object_get_string(r->pool, j_conf, "auth_request_params",
 			&provider->auth_request_params, cfg->provider.auth_request_params);
@@ -1065,6 +1139,11 @@ apr_byte_t oidc_metadata_conf_parse(request_rec *r, oidc_cfg *cfg,
 	oidc_json_object_get_string(r->pool, j_conf, "registration_token",
 			&provider->registration_token, cfg->provider.registration_token);
 
+	/* see if we've got custom registration request parameter values */
+	oidc_json_object_get_string(r->pool, j_conf, "registration_endpoint_json",
+			&provider->registration_endpoint_json,
+			cfg->provider.registration_endpoint_json);
+
 	/* get the flow to use */
 	oidc_json_object_get_string(r->pool, j_conf, "response_type",
 			&provider->response_type, NULL);
@@ -1135,7 +1214,7 @@ apr_byte_t oidc_metadata_client_parse(request_rec *r, oidc_cfg *cfg,
  * contents from both provider metadata directory and client metadata directory
  */
 apr_byte_t oidc_metadata_get(request_rec *r, oidc_cfg *cfg, const char *issuer,
-		oidc_provider_t **provider) {
+		oidc_provider_t **provider, apr_byte_t allow_discovery) {
 
 	apr_byte_t rc = FALSE;
 
@@ -1152,7 +1231,8 @@ apr_byte_t oidc_metadata_get(request_rec *r, oidc_cfg *cfg, const char *issuer,
 	 * NB: order is important here
 	 */
 
-	if (oidc_metadata_provider_get(r, cfg, issuer, &j_provider) == FALSE)
+	if (oidc_metadata_provider_get(r, cfg, issuer, &j_provider,
+			allow_discovery) == FALSE)
 		goto end;
 	if (oidc_metadata_provider_parse(r, j_provider, *provider) == FALSE)
 		goto end;
@@ -1179,4 +1259,3 @@ end:
 
 	return rc;
 }
-
