@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -51,6 +51,7 @@
 #ifndef MOD_AUTH_OPENIDC_H_
 #define MOD_AUTH_OPENIDC_H_
 
+#include <stdint.h>
 #include <openssl/evp.h>
 #include <apr_uri.h>
 #include <apr_uuid.h>
@@ -102,6 +103,16 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_IDTOKEN_SESSION_KEY "id_token"
 /* key for storing the access_token in the session context */
 #define OIDC_ACCESSTOKEN_SESSION_KEY "access_token"
+/* key for storing the access_token expiry in the session context */
+#define OIDC_ACCESSTOKEN_EXPIRES_SESSION_KEY "access_token_expires"
+/* key for storing the refresh_token in the session context */
+#define OIDC_REFRESHTOKEN_SESSION_KEY "refresh_token"
+/* key for storing maximum session duration in the session context */
+#define OIDC_SESSION_EXPIRES_SESSION_KEY "session_expires"
+/* key for storing request state */
+#define OIDC_REQUEST_STATE_SESSION_KEY "request_state"
+/* key for storing the original URL */
+#define OIDC_REQUEST_ORIGINAL_URL "original_url"
 
 /* key for storing the session_state in the session context */
 #define OIDC_SESSION_STATE_SESSION_KEY "session_state"
@@ -130,6 +141,9 @@ APLOG_USE_MODULE(auth_openidc);
 /* value that indicates to use client cookie based session tracking */
 #define OIDC_SESSION_TYPE_22_CLIENT_COOKIE 1
 
+/* nonce bytes length */
+#define OIDC_PROTO_NONCE_LENGTH 32
+
 /* pass id_token as individual claims in headers (default) */
 #define OIDC_PASS_IDTOKEN_AS_CLAIMS     1
 /* pass id_token payload as JSON object in header*/
@@ -140,9 +154,14 @@ APLOG_USE_MODULE(auth_openidc);
 /* prefix of the cookie that binds the state in the authorization request/response to the browser */
 #define OIDCStateCookiePrefix  "mod_auth_openidc_state_"
 
+/* default prefix for information passed in HTTP headers */
+#define OIDC_DEFAULT_HEADER_PREFIX "OIDC_"
+
 /* the (global) key for the mod_auth_openidc related state that is stored in the request userdata context */
 #define OIDC_USERDATA_KEY "mod_auth_openidc_state"
 
+#define OIDC_USERDATA_ENV_KEY "mod_auth_openidc_env"
+
 /* input filter hook name */
 #define OIDC_UTIL_HTTP_SENDSTRING "OIDC_UTIL_HTTP_SENDSTRING"
 
@@ -152,6 +171,12 @@ APLOG_USE_MODULE(auth_openidc);
 /* defines for how long provider metadata will be cached */
 #define OIDC_CACHE_PROVIDER_METADATA_EXPIRY_DEFAULT 86400
 
+/* define the parameter value for the "logout" request that indicates a GET-style logout call from the OP */
+#define OIDC_GET_STYLE_LOGOUT_PARAM_VALUE "get"
+
+/* define the name of the cookie/parameter for CSRF protection */
+#define OIDC_CSRF_NAME "x_csrf"
+
 /* cache sections */
 #define OIDC_CACHE_SECTION_JTI "jti"
 #define OIDC_CACHE_SECTION_SESSION "session"
@@ -160,6 +185,16 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_CACHE_SECTION_ACCESS_TOKEN "access_token"
 #define OIDC_CACHE_SECTION_PROVIDER "provider"
 
+typedef enum {
+	AUTHENTICATE, PASS, RETURN401
+} unauthenticated_action;
+
+typedef struct oidc_jwks_uri_t {
+	const char *url;
+	int refresh_interval;
+	int ssl_validate_server;
+} oidc_jwks_uri_t;
+
 typedef struct oidc_provider_t {
 	char *metadata_url;
 	char *issuer;
@@ -180,12 +215,14 @@ typedef struct oidc_provider_t {
 	char *client_name;
 	char *client_contact;
 	char *registration_token;
+	char *registration_endpoint_json;
 	char *scope;
 	char *response_type;
 	char *response_mode;
 	int jwks_refresh_interval;
 	int idtoken_iat_slack;
 	char *auth_request_params;
+	int session_max_duration;
 
 	char *client_jwks_uri;
 	char *id_token_signed_response_alg;
@@ -196,13 +233,27 @@ typedef struct oidc_provider_t {
 	char *userinfo_encrypted_response_enc;
 } oidc_provider_t ;
 
+typedef struct oidc_remote_user_claim_t {
+	const char *claim_name;
+	const char *reg_exp;
+} oidc_remote_user_claim_t;
+
 typedef struct oidc_oauth_t {
 	int ssl_validate_server;
 	char *client_id;
 	char *client_secret;
-	char *validate_endpoint_url;
-	char *validate_endpoint_auth;
-	char *remote_user_claim;
+	char *introspection_endpoint_url;
+	char *introspection_endpoint_method;
+	char *introspection_endpoint_params;
+	char *introspection_endpoint_auth;
+	char *introspection_token_param_name;
+	char *introspection_token_expiry_claim_name;
+	char *introspection_token_expiry_claim_format;
+	int introspection_token_expiry_claim_required;
+	oidc_remote_user_claim_t remote_user_claim;
+	apr_hash_t *verify_shared_keys;
+	char *verify_jwks_uri;
+	apr_hash_t *verify_public_keys;
 } oidc_oauth_t;
 
 typedef struct oidc_cfg {
@@ -211,8 +262,6 @@ typedef struct oidc_cfg {
 
 	/* the redirect URI as configured with the OpenID Connect OP's that we talk to */
 	char *redirect_uri;
-	/* (optional) external OP discovery page */
-	char *discover_url;
 	/* (optional) default URL for 3rd-party initiated SSO */
 	char *default_sso_url;
 	/* (optional) default URL to go to after logout */
@@ -245,6 +294,13 @@ typedef struct oidc_cfg {
 	char *cache_memcache_servers;
 	/* cache_type = shm: size of the shared memory segment (cq. max number of cached entries) */
 	int cache_shm_size_max;
+	/* cache_type = shm: maximum size in bytes of a cache entry */
+	int cache_shm_entry_size_max;
+#ifdef USE_LIBHIREDIS
+	/* cache_type= redis: Redis host/port server to use */
+	char *cache_redis_server;
+	char *cache_redis_password;
+#endif
 
 	/* tell the module to strip any mod_auth_openidc related headers that already have been set by the user-agent, normally required for secure operation */
 	int scrub_request_headers;
@@ -257,7 +313,7 @@ typedef struct oidc_cfg {
 	char *cookie_domain;
 	char *claim_delimiter;
 	char *claim_prefix;
-	char *remote_user_claim;
+	oidc_remote_user_claim_t remote_user_claim;
 	int pass_idtoken_as;
 	int cookie_http_only;
 
@@ -270,9 +326,15 @@ typedef struct oidc_cfg {
 } oidc_cfg;
 
 typedef struct oidc_dir_cfg {
+	/* (optional) external OP discovery page */
+	char *discover_url;
 	char *cookie_path;
 	char *cookie;
 	char *authn_header;
+	unauthenticated_action unauth_action;
+	apr_array_header_t *pass_cookies;
+	apr_byte_t pass_info_in_headers;
+	apr_byte_t pass_info_in_env_vars;
 } oidc_dir_cfg;
 
 int oidc_check_user_id(request_rec *r);
@@ -283,41 +345,40 @@ int oidc_auth_checker(request_rec *r);
 #endif
 void oidc_request_state_set(request_rec *r, const char *key, const char *value);
 const char*oidc_request_state_get(request_rec *r, const char *key);
+int oidc_handle_jwks(request_rec *r, oidc_cfg *c);
 
 // oidc_oauth
 int oidc_oauth_check_userid(request_rec *r, oidc_cfg *c);
 
 // oidc_proto.c
 
-typedef struct oidc_proto_state {
-	const char *nonce;
-	const char *original_url;
-	const char *original_method;
-	const char *issuer;
-	const char *response_type;
-	const char *response_mode;
-	const char *prompt;
-	apr_time_t timestamp;
-} oidc_proto_state;
-
-int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, oidc_proto_state *proto_state, const char *id_token_hint, const char *auth_request_params);
+int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, json_t *proto_state, const char *id_token_hint, const char *auth_request_params);
 apr_byte_t oidc_proto_is_post_authorization_response(request_rec *r, oidc_cfg *cfg);
 apr_byte_t oidc_proto_is_redirect_authorization_response(request_rec *r, oidc_cfg *cfg);
-apr_byte_t oidc_proto_check_token_type(request_rec *r, oidc_provider_t *provider, const char *token_type);
-apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *code, char **s_id_token, char **s_access_token, char **s_token_type);
-apr_byte_t oidc_proto_resolve_userinfo(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *access_token, const char **response, json_t **claims);
+apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *code, char **id_token, char **access_token, char **token_type, int *expires_in, char **refresh_token);
+apr_byte_t oidc_proto_refresh_request(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *rtoken, char **id_token, char **access_token, char **token_type, int *expires_in, char **refresh_token);
+apr_byte_t oidc_proto_resolve_userinfo(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *access_token, const char **response);
 apr_byte_t oidc_proto_account_based_discovery(request_rec *r, oidc_cfg *cfg, const char *acct, char **issuer);
-apr_byte_t oidc_proto_parse_idtoken(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token, const char *nonce, char **user, apr_jwt_t **jwt, apr_byte_t is_code_flow);
-apr_byte_t oidc_proto_validate_access_token(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *access_token, const char *token_type);
-apr_byte_t oidc_proto_validate_code(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *code);
+apr_byte_t oidc_proto_parse_idtoken(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token, const char *nonce, apr_jwt_t **jwt, apr_byte_t is_code_flow);
 int oidc_proto_javascript_implicit(request_rec *r, oidc_cfg *c);
 apr_array_header_t *oidc_proto_supported_flows(apr_pool_t *pool);
 apr_byte_t oidc_proto_flow_is_supported(apr_pool_t *pool, const char *flow);
 apr_byte_t oidc_proto_validate_authorization_response(request_rec *r, const char *response_type, const char *requested_response_mode, char **code, char **id_token, char **access_token, char **token_type, const char *used_response_mode);
-apr_byte_t oidc_proto_validate_code_response(request_rec *r, const char *response_type, char **id_token, char **access_token, char **token_type);
-apr_byte_t oidc_proto_idtoken_verify_signature(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, apr_jwt_t *jwt, apr_byte_t *refresh);
-apr_byte_t oidc_proto_validate_iat(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt);
-apr_byte_t oidc_proto_validate_exp(request_rec *r, apr_jwt_t *jwt);
+apr_byte_t oidc_proto_jwt_verify(request_rec *r, oidc_cfg *cfg, apr_jwt_t *jwt, const oidc_jwks_uri_t *jwks_uri, apr_hash_t *symmetric_keys);
+apr_byte_t oidc_proto_validate_jwt(request_rec *r, apr_jwt_t *jwt, const char *iss, apr_byte_t exp_is_mandatory, apr_byte_t iat_is_mandatory, int iat_slack);
+apr_byte_t oidc_proto_generate_nonce(request_rec *r, char **nonce, int len);
+
+apr_byte_t oidc_proto_authorization_response_code_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_authorization_response_code_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_code_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_code(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+
+// non-static for test.c
+apr_byte_t oidc_proto_validate_access_token(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *access_token);
+apr_byte_t oidc_proto_validate_code(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *code);
+apr_byte_t oidc_proto_validate_nonce(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *nonce, apr_jwt_t *jwt);
 
 // oidc_authz.c
 int oidc_authz_worker(request_rec *r, const json_t *const claims, const require_line *const reqs, int nelts);
@@ -335,7 +396,7 @@ void oidc_register_hooks(apr_pool_t *pool);
 // oidc_util.c
 int oidc_strnenvcmp(const char *a, const char *b, int len);
 int oidc_base64url_encode(request_rec *r, char **dst, const char *src, int src_len, int remove_padding);
-int oidc_base64url_decode(request_rec *r, char **dst, const char *src, int add_padding);
+int oidc_base64url_decode(request_rec *r, char **dst, const char *src);
 int oidc_encrypt_base64url_encode_string(request_rec *r, char **dst, const char *src);
 int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst, const char *src);
 char *oidc_get_current_url(const request_rec *r, const oidc_cfg *c);
@@ -344,31 +405,36 @@ char *oidc_normalize_header_name(const request_rec *r, const char *str);
 
 void oidc_util_set_cookie(request_rec *r, const char *cookieName, const char *cookieValue, apr_time_t expires);
 char *oidc_util_get_cookie(request_rec *r, const char *cookieName);
-apr_byte_t oidc_util_http_get(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
-apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
-apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url, const json_t *data, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
+apr_byte_t oidc_util_http_get(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies);
+apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies);
+apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url, const json_t *data, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies);
 apr_byte_t oidc_util_request_matches_url(request_rec *r, const char *url);
 apr_byte_t oidc_util_request_has_parameter(request_rec *r, const char* param);
 apr_byte_t oidc_util_get_request_parameter(request_rec *r, char *name, char **value);
 apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r, const char *str, json_t **json);
 int oidc_util_http_send(request_rec *r, const char *data, int data_len, const char *content_type, int success_rvalue);
-int oidc_util_html_send(request_rec *r, const char *html, int success_rvalue);
+int oidc_util_html_send(request_rec *r, const char *title, const char *html_head, const char *on_load, const char *html_body, int status_code);
 char *oidc_util_escape_string(const request_rec *r, const char *str);
 char *oidc_util_unescape_string(const request_rec *r, const char *str);
-apr_byte_t oidc_util_read_post(request_rec *r, apr_table_t *table);
-apr_byte_t oidc_util_generate_random_base64url_encoded_value(request_rec *r, int randomLen, char **randomB64);
+apr_byte_t oidc_util_read_form_encoded_params(request_rec *r, apr_table_t *table, const char *data);
+apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table);
 apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result);
 apr_byte_t oidc_util_issuer_match(const char *a, const char *b);
 int oidc_util_html_send_error(request_rec *r, const char *error, const char *description, int status_code);
 apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack, const char *needle);
-void oidc_util_set_app_header(request_rec *r, const char *s_key, const char *s_value, const char *claim_prefix);
-void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char *claim_prefix, const char *claim_delimiter);
+void oidc_util_set_app_info(request_rec *r, const char *s_key, const char *s_value, const char *claim_prefix, apr_byte_t as_header, apr_byte_t as_env_var);
+void oidc_util_set_app_infos(request_rec *r, const json_t *j_attrs, const char *claim_prefix, const char *claim_delimiter, apr_byte_t as_header, apr_byte_t as_env_var);
 apr_hash_t *oidc_util_spaced_string_to_hashtable(apr_pool_t *pool, const char *str);
 apr_byte_t oidc_util_spaced_string_equals(apr_pool_t *pool, const char *a, const char *b);
 apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool, const char *response_type, const char *match);
 apr_byte_t oidc_json_object_get_string(apr_pool_t *pool, json_t *json, const char *name, char **value, const char *default_value);
 apr_byte_t oidc_json_object_get_int(apr_pool_t *pool, json_t *json, const char *name, int *value, const int default_value);
 char *oidc_util_html_escape(apr_pool_t *pool, const char *input);
+void oidc_util_table_add_query_encoded_params(apr_pool_t *pool, apr_table_t *table, const char *params);
+apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *private_keys, const char *secret, const char *hash_algo);
+apr_hash_t * oidc_util_merge_key_sets(apr_pool_t *pool, apr_hash_t *k1, apr_hash_t *k2);
+apr_byte_t oidc_util_regexp_first_match(apr_pool_t *pool, const char *input, const char *regexp, char **output, char **error_str);
+apr_byte_t oidc_util_json_merge(json_t *src, json_t *dst);
 
 // oidc_crypto.c
 unsigned char *oidc_crypto_aes_encrypt(request_rec *r, oidc_cfg *cfg, unsigned char *plaintext, int *len);
@@ -379,13 +445,14 @@ apr_byte_t oidc_crypto_destroy(oidc_cfg *cfg, server_rec *s);
 apr_byte_t oidc_metadata_provider_retrieve(request_rec *r, oidc_cfg *cfg, const char *issuer, const char *url, json_t **j_metadata, const char **response);
 apr_byte_t oidc_metadata_provider_parse(request_rec *r, json_t *j_provider, oidc_provider_t *provider);
 apr_byte_t oidc_metadata_list(request_rec *r, oidc_cfg *cfg, apr_array_header_t **arr);
-apr_byte_t oidc_metadata_get(request_rec *r, oidc_cfg *cfg, const char *selected, oidc_provider_t **provider);
-apr_byte_t oidc_metadata_jwks_get(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, json_t **j_jwks, apr_byte_t *refresh);
+apr_byte_t oidc_metadata_get(request_rec *r, oidc_cfg *cfg, const char *selected, oidc_provider_t **provider, apr_byte_t allow_discovery);
+apr_byte_t oidc_metadata_jwks_get(request_rec *r, oidc_cfg *cfg, const oidc_jwks_uri_t *jwks_uri, json_t **j_jwks, apr_byte_t *refresh);
 
 // oidc_session.c
-#if MODULE_MAGIC_NUMBER_MAJOR >= 20081201
+#if MODULE_MAGIC_NUMBER_MAJOR_NOT_WORKING_YET >= 20081201
 // this stuff should make it easy to migrate to the post 2.3 mod_session infrastructure
 #include "mod_session.h"
+//#define OIDC_SESSION_USE_APACHE_SESSIONS 1
 #else
 typedef struct {
     apr_pool_t *pool;             /* pool to be used for this session */
