@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -80,6 +80,7 @@ apr_status_t oidc_session_load(request_rec *r, session_rec **zz) {
 	(*zz)->remote_user = apr_table_get((*zz)->entries,
 			OIDC_SESSION_REMOTE_USER_KEY);
 	const char *uuid = apr_table_get((*zz)->entries, OIDC_SESSION_UUID_KEY);
+	oidc_debug(r, "%s", uuid ? uuid : "<null>");
 	if (uuid != NULL)
 		apr_uuid_parse((*zz)->uuid, uuid);
 	return rc;
@@ -89,6 +90,7 @@ apr_status_t oidc_session_save(request_rec *r, session_rec *z) {
 	oidc_session_set(r, z, OIDC_SESSION_REMOTE_USER_KEY, z->remote_user);
 	char key[APR_UUID_FORMATTED_LENGTH + 1];
 	apr_uuid_format((char *) &key, z->uuid);
+	oidc_debug(r, "%s", key);
 	oidc_session_set(r, z, OIDC_SESSION_UUID_KEY, key);
 	return ap_session_save_fn(r, z);
 }
@@ -170,6 +172,7 @@ static char x2c(const char *what) {
 	return (digit);
 }
 
+#ifndef WIN32
 AP_DECLARE(char *) ap_escape_urlencoded_buffer(char *copy, const char *buffer) {
 	const unsigned char *s = (const unsigned char *) buffer;
 	unsigned char *d = (unsigned char *) copy;
@@ -188,6 +191,7 @@ AP_DECLARE(char *) ap_escape_urlencoded_buffer(char *copy, const char *buffer) {
 	*d = '\0';
 	return copy;
 }
+#endif
 
 static int oidc_session_unescape_url(char *url, const char *forbid,
 		const char *reserved) {
@@ -238,6 +242,7 @@ static int oidc_session_unescape_url(char *url, const char *forbid,
 	}
 }
 
+#ifndef WIN32
 AP_DECLARE(int) ap_unescape_urlencoded(char *query) {
 	char *slider;
 	/* replace plus with a space */
@@ -251,6 +256,7 @@ AP_DECLARE(int) ap_unescape_urlencoded(char *query) {
 	/* unescape everything else */
 	return oidc_session_unescape_url(query, NULL, NULL);
 }
+#endif
 
 // copied from mod_session.c
 static apr_status_t oidc_session_identity_decode(request_rec * r,
@@ -259,7 +265,7 @@ static apr_status_t oidc_session_identity_decode(request_rec * r,
 	char *encoded, *pair;
 	const char *sep = "&";
 
-	oidc_debug(r, "decoding %s", z->encoded);
+	//oidc_debug(r, "decoding %s", z->encoded);
 
 	/* sanity check - anything to decode? */
 	if (!z->encoded) {
@@ -275,7 +281,7 @@ static apr_status_t oidc_session_identity_decode(request_rec * r,
 		char *key = apr_strtok(pair, psep, &plast);
 		char *val = apr_strtok(NULL, psep, &plast);
 
-		oidc_debug(r, "decoding %s=%s", key, val);
+		//oidc_debug(r, "decoding %s=%s", key, val);
 
 		if (key && *key) {
 			if (!val || !*val) {
@@ -351,13 +357,11 @@ static apr_status_t oidc_session_load_cache(request_rec *r, session_rec *z) {
 
 	/* get the string-encoded session from the cache based on the key */
 	if (uuid != NULL) {
-		if (c->cache->get(r, OIDC_CACHE_SECTION_SESSION, uuid,
-				&z->encoded) == TRUE)
-			return APR_SUCCESS;
+		c->cache->get(r, OIDC_CACHE_SECTION_SESSION, uuid, &z->encoded);
 		//oidc_util_set_cookie(r, d->cookie, "");
 	}
 
-	return APR_EGENERAL;
+	return (z->encoded != NULL) ? APR_SUCCESS : APR_EGENERAL;
 }
 
 /*
@@ -416,7 +420,10 @@ static apr_status_t oidc_session_save_cookie(request_rec *r, session_rec *z) {
 
 	char *cookieValue = "";
 	if (z->encoded && z->encoded[0]) {
-		oidc_encrypt_base64url_encode_string(r, &cookieValue, z->encoded);
+		if (oidc_encrypt_base64url_encode_string(r, &cookieValue, z->encoded) <= 0) {
+			oidc_error(r, "oidc_encrypt_base64url_encode_string failed");
+			return APR_EGENERAL;
+		}
 	}
 	oidc_util_set_cookie(r, d->cookie, cookieValue, -1);
 
@@ -475,6 +482,9 @@ static apr_status_t oidc_session_load_22(request_rec *r, session_rec **zz) {
 
 		oidc_warn(r, "session restored from cache has expired");
 		apr_table_clear(z->entries);
+		z->expiry = 0;
+		z->encoded = NULL;
+
 		return APR_EGENERAL;
 	}
 
