@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2014 Ping Identity Corporation
+ * Copyright (C) 2013-2015 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -62,6 +62,8 @@
 
 #include "mod_auth_openidc.h"
 
+#include <pcre.h>
+
 /* hrm, should we get rid of this by adding parameters to the (3) functions? */
 extern module AP_MODULE_DECLARE_DATA auth_openidc_module;
 
@@ -103,8 +105,7 @@ int oidc_base64url_encode(request_rec *r, char **dst, const char *src,
 /*
  * base64url decode a string
  */
-int oidc_base64url_decode(request_rec *r, char **dst, const char *src,
-		int add_padding) {
+int oidc_base64url_decode(request_rec *r, char **dst, const char *src) {
 	if (src == NULL) {
 		oidc_error(r, "not decoding anything; src=NULL");
 		return -1;
@@ -120,19 +121,17 @@ int oidc_base64url_decode(request_rec *r, char **dst, const char *src,
 			dec[i] = '=';
 		i++;
 	}
-	if (add_padding == 1) {
-		switch (strlen(dec) % 4) {
-		case 0:
-			break;
-		case 2:
-			dec = apr_pstrcat(r->pool, dec, "==", NULL);
-			break;
-		case 3:
-			dec = apr_pstrcat(r->pool, dec, "=", NULL);
-			break;
-		default:
-			return 0;
-		}
+	switch (strlen(dec) % 4) {
+	case 0:
+		break;
+	case 2:
+		dec = apr_pstrcat(r->pool, dec, "==", NULL);
+		break;
+	case 3:
+		dec = apr_pstrcat(r->pool, dec, "=", NULL);
+		break;
+	default:
+		return 0;
 	}
 	int dlen = apr_base64_decode_len(dec);
 	*dst = apr_palloc(r->pool, dlen);
@@ -164,7 +163,7 @@ int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst,
 	oidc_cfg *c = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
 	char *decbuf = NULL;
-	int dec_len = oidc_base64url_decode(r, &decbuf, src, 1);
+	int dec_len = oidc_base64url_decode(r, &decbuf, src);
 	if (dec_len <= 0) {
 		oidc_error(r, "oidc_base64url_decode failed");
 		return -1;
@@ -272,7 +271,7 @@ char *oidc_util_unescape_string(const request_rec *r, const char *str) {
 char *oidc_util_html_escape(apr_pool_t *pool, const char *s) {
 	const char chars[6] = { '&', '\'', '\"', '>', '<', '\0' };
 	const char * const replace[] =
-	{ "&amp;", "&apos;", "&quot;", "&gt;", "&lt;", };
+			{ "&amp;", "&apos;", "&quot;", "&gt;", "&lt;", };
 	unsigned int i, j = 0, k, n = 0, len = strlen(chars);
 	int m = 0;
 	char *r = apr_pcalloc(pool, strlen(s) * 6);
@@ -295,7 +294,6 @@ char *oidc_util_html_escape(apr_pool_t *pool, const char *s) {
 	return apr_pstrdup(pool, r);
 }
 
-
 /*
  * get the URL scheme that is currently being accessed
  */
@@ -414,11 +412,13 @@ static int oidc_http_add_form_url_encoded_param(void* rec, const char* key,
 static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 		const char *data, const char *content_type, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
-		const char **response, int timeout, const char *outgoing_proxy) {
+		const char **response, int timeout, const char *outgoing_proxy,
+		apr_array_header_t *pass_cookies) {
 	char curlError[CURL_ERROR_SIZE];
 	oidc_curl_buffer curlBuffer;
 	CURL *curl;
 	struct curl_slist *h_list = NULL;
+	int i;
 
 	/* do some logging about the inputs */
 	oidc_debug(r,
@@ -461,6 +461,19 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 	curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST,
 			(ssl_validate_server != FALSE ? 2L : 0L));
 
+#ifdef WIN32
+	DWORD buflen;
+	char *ptr = NULL;
+	char *retval = (char *) malloc(sizeof (TCHAR) * (MAX_PATH + 1));
+	retval[0] = '\0';
+	buflen = SearchPath(NULL, "curl-ca-bundle.crt", NULL, MAX_PATH+1, retval, &ptr);
+	if (buflen > 0)
+		curl_easy_setopt(curl, CURLOPT_CAINFO, retval);
+	else
+		oidc_warn(r, "no curl-ca-bundle.crt file found in path");
+	free(retval);
+#endif
+
 	/* identify this HTTP client */
 	curl_easy_setopt(curl, CURLOPT_USERAGENT, "mod_auth_openidc");
 
@@ -499,6 +512,28 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 	if (h_list != NULL)
 		curl_easy_setopt(curl, CURLOPT_HTTPHEADER, h_list);
 
+	/* gather cookies that we need to pass on from the incoming request */
+	char *cookie_string = NULL;
+	for (i = 0; i < pass_cookies->nelts; i++) {
+		const char *cookie_name = ((const char**) pass_cookies->elts)[i];
+		char *cookie_value = oidc_util_get_cookie(r, cookie_name);
+		if (cookie_value != NULL) {
+			cookie_string =
+					(cookie_string == NULL) ?
+							apr_psprintf(r->pool, "%s=%s", cookie_name,
+									cookie_value) :
+									apr_psprintf(r->pool, "%s; %s=%s", cookie_string,
+											cookie_name, cookie_value);
+		}
+	}
+
+	/* see if we need to pass any cookies */
+	if (cookie_string != NULL) {
+		oidc_debug(r, "passing browser cookies on backend call: %s",
+				cookie_string);
+		curl_easy_setopt(curl, CURLOPT_COOKIE, cookie_string);
+	}
+
 	/* set the target URL */
 	curl_easy_setopt(curl, CURLOPT_URL, url);
 
@@ -515,7 +550,7 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 	/* set and log the response */
 	oidc_debug(r, "response=%s", *response);
 
-	out:
+out:
 
 	/* cleanup and return the result */
 	if (h_list != NULL)
@@ -531,7 +566,8 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 apr_byte_t oidc_util_http_get(request_rec *r, const char *url,
 		const apr_table_t *params, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
-		const char **response, int timeout, const char *outgoing_proxy) {
+		const char **response, int timeout, const char *outgoing_proxy,
+		apr_array_header_t *pass_cookies) {
 
 	if ((params != NULL) && (apr_table_elts(params)->nelts > 0)) {
 		oidc_http_encode_t data = { r, "" };
@@ -542,7 +578,8 @@ apr_byte_t oidc_util_http_get(request_rec *r, const char *url,
 	}
 
 	return oidc_util_http_call(r, url, NULL, NULL, basic_auth, bearer_token,
-			ssl_validate_server, response, timeout, outgoing_proxy);
+			ssl_validate_server, response, timeout, outgoing_proxy,
+			pass_cookies);
 }
 
 /*
@@ -551,7 +588,8 @@ apr_byte_t oidc_util_http_get(request_rec *r, const char *url,
 apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
 		const apr_table_t *params, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
-		const char **response, int timeout, const char *outgoing_proxy) {
+		const char **response, int timeout, const char *outgoing_proxy,
+		apr_array_header_t *pass_cookies) {
 
 	const char *data = NULL;
 	if ((params != NULL) && (apr_table_elts(params)->nelts > 0)) {
@@ -564,7 +602,8 @@ apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
 
 	return oidc_util_http_call(r, url, data,
 			"application/x-www-form-urlencoded", basic_auth, bearer_token,
-			ssl_validate_server, response, timeout, outgoing_proxy);
+			ssl_validate_server, response, timeout, outgoing_proxy,
+			pass_cookies);
 }
 
 /*
@@ -573,7 +612,7 @@ apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
 apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url,
 		const json_t *json, const char *basic_auth, const char *bearer_token,
 		int ssl_validate_server, const char **response, int timeout,
-		const char *outgoing_proxy) {
+		const char *outgoing_proxy, apr_array_header_t *pass_cookies) {
 
 	char *data = NULL;
 	if (json != NULL) {
@@ -584,7 +623,7 @@ apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url,
 
 	return oidc_util_http_call(r, url, data, "application/json", basic_auth,
 			bearer_token, ssl_validate_server, response, timeout,
-			outgoing_proxy);
+			outgoing_proxy, pass_cookies);
 }
 
 /*
@@ -647,15 +686,16 @@ void oidc_util_set_cookie(request_rec *r, const char *cookieName,
 	}
 
 	/* construct the cookie value */
-	headerString = apr_psprintf(r->pool, "%s=%s;%s;Path=%s%s%s%s", cookieName,
+	headerString = apr_psprintf(r->pool, "%s=%s;Path=%s%s%s%s%s", cookieName,
 			cookieValue,
-			((apr_strnatcasecmp("https", oidc_get_current_url_scheme(r)) == 0) ?
-					";Secure" : ""), oidc_util_get_cookie_path(r),
+			oidc_util_get_cookie_path(r),
+			(expiresString == NULL) ?
+					"" : apr_psprintf(r->pool, "; expires=%s", expiresString),
 			c->cookie_domain != NULL ?
 					apr_psprintf(r->pool, ";Domain=%s", c->cookie_domain) : "",
-			c->cookie_http_only != FALSE ? ";HttpOnly" : "",
-			(expiresString == NULL) ?
-					"" : apr_psprintf(r->pool, "; expires=%s", expiresString));
+			((apr_strnatcasecmp("https", oidc_get_current_url_scheme(r)) == 0) ?
+					";Secure" : ""),
+			c->cookie_http_only != FALSE ? ";HttpOnly" : "");
 
 	/* use r->err_headers_out so we always print our headers (even on 302 redirect) - headers_out only prints on 2xx responses */
 	apr_table_add(r->err_headers_out, "Set-Cookie", headerString);
@@ -709,7 +749,7 @@ char *oidc_util_get_cookie(request_rec *r, const char *cookieName) {
 	}
 
 	/* log what we've found */
-	oidc_debug(r, "returning %s", rv);
+	oidc_debug(r, "returning \"%s\" = %s", cookieName, rv ? apr_psprintf(r->pool, "\"%s\"", rv) : "<null>");
 
 	return rv;
 }
@@ -748,11 +788,10 @@ apr_byte_t oidc_util_request_matches_url(request_rec *r, const char *url) {
 	apr_uri_t uri;
 	memset(&uri, 0, sizeof(apr_uri_t));
 	apr_uri_parse(r->pool, url, &uri);
-	apr_byte_t rc =
-			(apr_strnatcmp(r->parsed_uri.path, uri.path) == 0) ? TRUE : FALSE;
-	oidc_debug(r, "comparing \"%s\"==\"%s\" (%d)", r->parsed_uri.path, uri.path,
-			rc);
-	return rc;
+	oidc_debug(r, "comparing \"%s\"==\"%s\"", r->parsed_uri.path, uri.path);
+	if ((r->parsed_uri.path == NULL) || (uri.path == NULL))
+		return (r->parsed_uri.path == uri.path);
+	return (apr_strnatcmp(r->parsed_uri.path, uri.path) == 0);
 }
 
 /*
@@ -772,7 +811,6 @@ apr_byte_t oidc_util_request_has_parameter(request_rec *r, const char* param) {
  */
 apr_byte_t oidc_util_get_request_parameter(request_rec *r, char *name,
 		char **value) {
-	// TODO: we should really check with ? and & and avoid any <bogus>code= stuff to trigger true
 	char *tokenizer_ctx, *p, *args;
 	const char *k_param = apr_psprintf(r->pool, "%s=", name);
 	const size_t k_param_sz = strlen(k_param);
@@ -803,16 +841,11 @@ apr_byte_t oidc_util_get_request_parameter(request_rec *r, char *name,
 static apr_byte_t oidc_util_json_string_print(request_rec *r, json_t *result,
 		const char *key, const char *log) {
 	json_t *value = json_object_get(result, key);
-	if (value != NULL) {
-		if (json_is_string(value)) {
-			oidc_error(r,
-					"%s: response contained a \"%s\" key with string value: \"%s\"",
-					log, key, json_string_value(value));
-		} else {
-			oidc_error(r,
-					"%s: response contained an \"%s\" key but no string value",
-					log, key);
-		}
+	if (value != NULL && !json_is_null(value)) {
+		char *s_value = json_dumps(value, JSON_ENCODE_ANY);
+		oidc_error(r, "%s: response contained an \"%s\" entry with value: \"%s\"",
+				log, key, s_value);
+		free(s_value);
 		return TRUE;
 	}
 	return FALSE;
@@ -868,8 +901,8 @@ apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
 /*
  * sends content to the user agent
  */
-int oidc_util_http_send(request_rec *r, const char *data, int data_len, const char *content_type,
-		int success_rvalue) {
+int oidc_util_http_send(request_rec *r, const char *data, int data_len,
+		const char *content_type, int success_rvalue) {
 	ap_set_content_type(r, content_type);
 	apr_bucket_brigade *bb = apr_brigade_create(r->pool,
 			r->connection->bucket_alloc);
@@ -885,11 +918,52 @@ int oidc_util_http_send(request_rec *r, const char *data, int data_len, const ch
 }
 
 /*
- * sends HTML content to the user agent
+ * send HTML content to the user agent
  */
-int oidc_util_html_send(request_rec *r, const char *html,
-		int success_rvalue) {
-	return oidc_util_http_send(r, html, strlen(html), "text/html", success_rvalue);
+int oidc_util_html_send(request_rec *r, const char *title,
+		const char *html_head, const char *on_load, const char *html_body,
+		int status_code) {
+
+	char *html =
+			"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n"
+					"<html>\n"
+					"  <head>\n"
+					"    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
+					"    <title>%s</title>\n"
+					"    %s\n"
+					"  </head>\n"
+					"  <body%s>\n"
+					"%s\n"
+					"  </body>\n"
+					"</html>\n";
+
+	html = apr_psprintf(r->pool, html,
+			title ? oidc_util_html_escape(r->pool, title) : "",
+			html_head ? html_head : "",
+			on_load ? apr_psprintf(r->pool, " onload=\"%s()\"", on_load) : "",
+			html_body ? html_body : "<p></p>");
+
+	return oidc_util_http_send(r, html, strlen(html), "text/html", status_code);
+}
+
+/*
+ * send a user-facing error to the browser
+ */
+int oidc_util_html_send_error(request_rec *r, const char *error,
+		const char *description, int status_code) {
+
+	char *html_body = "";
+
+	if (error != NULL) {
+		html_body = apr_psprintf(r->pool, "%s<p>Error: <pre>%s</pre></p>",
+				html_body, oidc_util_html_escape(r->pool, error));
+	}
+	if (description != NULL) {
+		html_body = apr_psprintf(r->pool, "%s<p>Description: <pre>%s</pre></p>",
+				html_body, oidc_util_html_escape(r->pool, description));
+	}
+
+	return oidc_util_html_send(r, "Error", NULL, NULL, html_body, status_code);
 }
 
 /*
@@ -922,12 +996,31 @@ static apr_byte_t oidc_util_read(request_rec *r, const char **rbuf) {
 	return TRUE;
 }
 
+/*
+ * read form-encoded parameters from a string in to a table
+ */
+apr_byte_t oidc_util_read_form_encoded_params(request_rec *r,
+		apr_table_t *table, const char *data) {
+	const char *key, *val, *p = data;
+
+	while (p && *p && (val = ap_getword(r->pool, &p, '&'))) {
+		key = ap_getword(r->pool, &val, '=');
+		key = oidc_util_unescape_string(r, key);
+		val = oidc_util_unescape_string(r, val);
+		apr_table_set(table, key, val);
+	}
+
+	oidc_debug(r, "parsed: \"%s\" in to %d elements", data,
+			apr_table_elts(table)->nelts);
+
+	return TRUE;
+}
+
 /*
  * read the POST parameters in to a table
  */
-apr_byte_t oidc_util_read_post(request_rec *r, apr_table_t *table) {
+apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table) {
 	const char *data = NULL;
-	const char *key, *val;
 
 	if (r->method_number != M_POST)
 		return FALSE;
@@ -935,37 +1028,7 @@ apr_byte_t oidc_util_read_post(request_rec *r, apr_table_t *table) {
 	if (oidc_util_read(r, &data) != TRUE)
 		return FALSE;
 
-	while (data && *data && (val = ap_getword(r->pool, &data, '&'))) {
-		key = ap_getword(r->pool, &val, '=');
-		key = oidc_util_unescape_string(r, key);
-		val = oidc_util_unescape_string(r, val);
-		//ap_unescape_url((char*) key);
-		//ap_unescape_url((char*) val);
-		apr_table_set(table, key, val);
-	}
-
-	return TRUE;
-}
-
-// TODO: check return values
-apr_byte_t oidc_util_generate_random_base64url_encoded_value(request_rec *r,
-		int randomLen, char **randomB64) {
-	unsigned char *brnd = apr_pcalloc(r->pool, randomLen);
-	apr_generate_random_bytes((unsigned char *) brnd, randomLen);
-	*randomB64 = apr_palloc(r->pool, apr_base64_encode_len(randomLen) + 1);
-	char *enc = *randomB64;
-	apr_base64_encode(enc, (const char *) brnd, randomLen);
-	int i = 0;
-	while (enc[i] != '\0') {
-		if (enc[i] == '+')
-			enc[i] = '-';
-		if (enc[i] == '/')
-			enc[i] = '_';
-		if (enc[i] == '=')
-			enc[i] = ',';
-		i++;
-	}
-	return TRUE;
+	return oidc_util_read_form_encoded_params(r, table, data);
 }
 
 /*
@@ -1030,7 +1093,7 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 
 	return TRUE;
 
-	error_close:
+error_close:
 
 	apr_file_unlock(fd);
 	apr_file_close(fd);
@@ -1060,29 +1123,6 @@ apr_byte_t oidc_util_issuer_match(const char *a, const char *b) {
 	return TRUE;
 }
 
-/*
- * send a user-facing error to the browser
- * TODO: more templating
- */
-int oidc_util_html_send_error(request_rec *r, const char *error,
-		const char *description, int status_code) {
-	char *msg =
-			"<html><body><p>the OpenID Connect Provider returned an error:</p>";
-
-	if (error != NULL) {
-		msg = apr_psprintf(r->pool, "%s<p>Error: <pre>%s</pre></p>", msg,
-				oidc_util_html_escape(r->pool, error));
-	}
-	if (description != NULL) {
-		msg = apr_psprintf(r->pool, "%s<p>Description: <pre>%s</pre></p>", msg,
-				oidc_util_html_escape(r->pool, description));
-	}
-
-	msg = apr_psprintf(r->pool, "%s</body></html>", msg);
-
-	return oidc_util_html_send(r, msg, status_code);
-}
-
 /*
  * see if a certain string value is part of a JSON array with string elements
  */
@@ -1115,25 +1155,57 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack,
 /*
  * set an HTTP header to pass information to the application
  */
-void oidc_util_set_app_header(request_rec *r, const char *s_key,
-		const char *s_value, const char *claim_prefix) {
+void oidc_util_set_app_info(request_rec *r, const char *s_key,
+		const char *s_value, const char *claim_prefix, apr_byte_t as_header,
+		apr_byte_t as_env_var) {
+
+	apr_table_t *env = NULL;
 
 	/* construct the header name, cq. put the prefix in front of a normalized key name */
 	const char *s_name = apr_psprintf(r->pool, "%s%s", claim_prefix,
 			oidc_normalize_header_name(r, s_key));
 
-	/* do some logging about this event */
-	oidc_debug(r, "setting header \"%s: %s\"", s_name, s_value);
+	if (as_header) {
+		/*
+		 * sanitize the header value by replacing line feeds with spaces
+		 * just like the Apache header input algorithms do for incoming headers
+		 *
+		 * this makes it impossible to have line feeds in values but that is
+		 * compliant with RFC 7230 (and impossible for regular headers due to Apache's
+		 * parsing of headers anyway) and fixes a security vulnerability on
+		 * overwriting/setting outgoing headers when used in proxy mode
+		 */
+		char *p = NULL;
+		while ((p = strchr(s_value, '\n')))
+			*p = ' ';
 
-	/* now set the actual header name/value */
-	apr_table_set(r->headers_in, s_name, s_value);
+		/* do some logging about this event */
+		oidc_debug(r, "setting header \"%s: %s\"", s_name, s_value);
+
+		/* now set the actual header name/value */
+		apr_table_set(r->headers_in, s_name, s_value);
+	}
+
+	if (as_env_var) {
+
+		/* do some logging about this event */
+		oidc_debug(r, "setting environment variable \"%s: %s\"", s_name,
+				s_value);
+
+		apr_pool_userdata_get((void **) &env, OIDC_USERDATA_ENV_KEY, r->pool);
+		if (env == NULL)
+			env = apr_table_make(r->pool, 10);
+		apr_table_set(env, s_name, s_value);
+		apr_pool_userdata_set(env, OIDC_USERDATA_ENV_KEY, NULL, r->pool);
+	}
 }
 
 /*
  * set the user/claims information from the session in HTTP headers passed on to the application
  */
-void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
-		const char *claim_prefix, const char *claim_delimiter) {
+void oidc_util_set_app_infos(request_rec *r, const json_t *j_attrs,
+		const char *claim_prefix, const char *claim_delimiter,
+		apr_byte_t as_header, apr_byte_t as_env_var) {
 
 	char s_int[255];
 	json_t *j_value = NULL;
@@ -1153,29 +1225,31 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
 		s_key = json_object_iter_key(iter);
 		j_value = json_object_iter_value(iter);
 
-//		char *s_value= json_dumps(j_value, JSON_ENCODE_ANY);
-//		oidc_util_set_app_header(r, s_key, s_value, claim_prefix);
-//		free(s_value);
+		//		char *s_value= json_dumps(j_value, JSON_ENCODE_ANY);
+		//		oidc_util_set_app_info(r, s_key, s_value, claim_prefix);
+		//		free(s_value);
 
 		/* check if it is a single value string */
 		if (json_is_string(j_value)) {
 
 			/* set the single string in the application header whose name is based on the key and the prefix */
-			oidc_util_set_app_header(r, s_key, json_string_value(j_value),
-					claim_prefix);
+			oidc_util_set_app_info(r, s_key, json_string_value(j_value),
+					claim_prefix, as_header, as_env_var);
 
 		} else if (json_is_boolean(j_value)) {
 
 			/* set boolean value in the application header whose name is based on the key and the prefix */
-			oidc_util_set_app_header(r, s_key,
-			json_is_true(j_value) ? "1" : "0", claim_prefix);
+			oidc_util_set_app_info(r, s_key,
+					(json_is_true(j_value) ? "1" : "0"), claim_prefix,
+					as_header, as_env_var);
 
 		} else if (json_is_integer(j_value)) {
 
-			if (sprintf(s_int, "%" JSON_INTEGER_FORMAT,
-					json_integer_value(j_value)) > 0) {
+			if (sprintf(s_int, "%ld",
+					(long)json_integer_value(j_value)) > 0) {
 				/* set long value in the application header whose name is based on the key and the prefix */
-				oidc_util_set_app_header(r, s_key, s_int, claim_prefix);
+				oidc_util_set_app_info(r, s_key, s_int, claim_prefix, as_header,
+						as_env_var);
 			} else {
 				oidc_warn(r,
 						"could not convert JSON number to string (> 255 characters?), skipping");
@@ -1184,15 +1258,16 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
 		} else if (json_is_real(j_value)) {
 
 			/* set float value in the application header whose name is based on the key and the prefix */
-			oidc_util_set_app_header(r, s_key,
+			oidc_util_set_app_info(r, s_key,
 					apr_psprintf(r->pool, "%lf", json_real_value(j_value)),
-					claim_prefix);
+					claim_prefix, as_header, as_env_var);
 
 		} else if (json_is_object(j_value)) {
 
 			/* set json value in the application header whose name is based on the key and the prefix */
 			char *s_value = json_dumps(j_value, 0);
-			oidc_util_set_app_header(r, s_key, s_value, claim_prefix);
+			oidc_util_set_app_info(r, s_key, s_value, claim_prefix, as_header,
+					as_env_var);
 			free(s_value);
 
 			/* check if it is a multi-value string */
@@ -1200,8 +1275,8 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
 
 			/* some logging about what we're going to do */
 			oidc_debug(r,
-					"parsing attribute array for key \"%s\" (#nr-of-elems: %zu)",
-					s_key, json_array_size(j_value));
+					"parsing attribute array for key \"%s\" (#nr-of-elems: %llu)",
+					s_key, (unsigned long long )json_array_size(j_value));
 
 			/* string to hold the concatenated array string values */
 			char *s_concat = apr_pstrdup(r->pool, "");
@@ -1234,7 +1309,7 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
 								json_is_true(elem) ? "1" : "0");
 					} else {
 						s_concat = apr_psprintf(r->pool, "%s",
-						json_is_true(elem) ? "1" : "0");
+								json_is_true(elem) ? "1" : "0");
 					}
 
 				} else {
@@ -1247,7 +1322,8 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
 			}
 
 			/* set the concatenated string */
-			oidc_util_set_app_header(r, s_key, s_concat, claim_prefix);
+			oidc_util_set_app_info(r, s_key, s_concat, claim_prefix, as_header,
+					as_env_var);
 
 		} else {
 
@@ -1341,3 +1417,161 @@ apr_byte_t oidc_json_object_get_int(apr_pool_t *pool, json_t *json,
 	}
 	return TRUE;
 }
+
+/*
+ * merge two JSON objects
+ */
+ apr_byte_t oidc_util_json_merge(json_t *src, json_t *dst) {
+
+	const char *key;
+	json_t *value = NULL;
+	void *iter = NULL;
+
+	if ((src == NULL) || (dst == NULL)) return FALSE;
+
+	iter = json_object_iter(src);
+	while(iter) {
+		key = json_object_iter_key(iter);
+		value = json_object_iter_value(iter);
+		json_object_set(dst, key, value);
+		iter = json_object_iter_next(src, iter);
+	}
+
+	return TRUE;
+}
+
+/*
+ * add query encoded parameters to a table
+ */
+void oidc_util_table_add_query_encoded_params(apr_pool_t *pool,
+		apr_table_t *table, const char *params) {
+	if (params != NULL) {
+		const char *key, *val;
+		const char *p = params;
+		while (*p && (val = ap_getword(pool, &p, '&'))) {
+			key = ap_getword(pool, &val, '=');
+			ap_unescape_url((char *) key);
+			ap_unescape_url((char *) val);
+			apr_table_addn(table, key, val);
+		}
+	}
+}
+
+/*
+ * merge provided keys and client secret in to a single hashtable
+ */
+apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *keys,
+		const char *client_secret, const char *hash_algo) {
+	apr_jwt_error_t err;
+	apr_jwk_t *jwk = NULL;
+	unsigned char *key = NULL;
+	unsigned int key_len;
+	apr_hash_t *result = NULL;
+
+	result = (keys != NULL) ? apr_hash_copy(pool, keys) : apr_hash_make(pool);
+
+	if (client_secret != NULL) {
+
+		if (hash_algo == NULL) {
+			key = (unsigned char *) client_secret;
+			key_len = strlen(client_secret);
+		} else {
+			/* hash the client_secret first, this is OpenID Connect specific */
+			apr_jws_hash_bytes(pool, hash_algo,
+					(const unsigned char *) client_secret,
+					strlen(client_secret), &key, &key_len, &err);
+		}
+
+		if (apr_jwk_parse_symmetric_key(pool, key, key_len, &jwk, &err) == TRUE) {
+			apr_hash_set(result, jwk->kid, APR_HASH_KEY_STRING, jwk);
+		}
+	}
+
+	return result;
+}
+
+/*
+ * merge two key sets
+ */
+apr_hash_t * oidc_util_merge_key_sets(apr_pool_t *pool, apr_hash_t *k1,
+		apr_hash_t *k2) {
+	if (k1 == NULL) {
+		if (k2 == NULL)
+			return apr_hash_make(pool);
+		return k2;
+	}
+	if (k2 == NULL)
+		return k1;
+	return apr_hash_overlay(pool, k1, k2);
+}
+
+/*
+ * regexp match
+ */
+#define OIDC_UTIL_REGEXP_MATCH_SIZE 30
+#define OIDC_UTIL_REGEXP_MATCH_NR 1
+
+apr_byte_t oidc_util_regexp_first_match(apr_pool_t *pool, const char *input,
+		const char *regexp, char **output, char **error_str) {
+	const char *errorptr;
+	int erroffset;
+	pcre *preg;
+	int subStr[OIDC_UTIL_REGEXP_MATCH_SIZE];
+	const char *psubStrMatchStr;
+
+	preg = pcre_compile(regexp, 0, &errorptr, &erroffset, NULL);
+
+	if (preg == NULL) {
+		*error_str = apr_psprintf(pool,
+				"pattern [%s] is not a valid regular expression", regexp);
+		pcre_free(preg);
+		return FALSE;
+	}
+
+	int rc = 0;
+	if ((rc = pcre_exec(preg, NULL, input, (int) strlen(input), 0, 0, subStr,
+			OIDC_UTIL_REGEXP_MATCH_SIZE)) < 0) {
+		switch (rc) {
+		case PCRE_ERROR_NOMATCH:
+			*error_str = apr_pstrdup(pool, "string did not match the pattern");
+			break;
+		case PCRE_ERROR_NULL:
+			*error_str = apr_pstrdup(pool, "something was null");
+			break;
+		case PCRE_ERROR_BADOPTION:
+			*error_str = apr_pstrdup(pool, "a bad option was passed");
+			break;
+		case PCRE_ERROR_BADMAGIC:
+			*error_str = apr_pstrdup(pool,
+					"magic number bad (compiled re corrupt?)");
+			break;
+		case PCRE_ERROR_UNKNOWN_NODE:
+			*error_str = apr_pstrdup(pool,
+					"something kooky in the compiled re");
+			break;
+		case PCRE_ERROR_NOMEMORY:
+			*error_str = apr_pstrdup(pool, "ran out of memory");
+			break;
+		default:
+			*error_str = apr_psprintf(pool, "unknown error: %d", rc);
+			break;
+		}
+		pcre_free(preg);
+		return FALSE;
+	}
+
+	if (pcre_get_substring(input, subStr, rc, OIDC_UTIL_REGEXP_MATCH_NR,
+			&(psubStrMatchStr)) <= 0) {
+		*error_str = apr_psprintf(pool, "pcre_get_substring failed (rc=%d)",
+				rc);
+		pcre_free(preg);
+		return FALSE;
+	}
+
+	*output = apr_pstrdup(pool, psubStrMatchStr);
+
+	pcre_free_substring(psubStrMatchStr);
+	pcre_free(preg);
+
+	return TRUE;
+}
